import numpy as np
import pandas as pd
import datetime as dt
import scipy.stats as stats
import statsmodels
import warnings

from .make_figures import *


def synthesize_data(n: int,
                    data: pd.DataFrame,
                    precip_dict: dict,
                    copulaetemp_dict: dict,
                    resolution: str,
                    validate: bool,
                    dirpath: str,
                    synthesize_kwargs: dict) -> pd.DataFrame:
    """
    Managing function that synthesizes weather from the fit or given 
    statistical parameters 

    Parameters
    ----------
    n: int
        Number of years for the generator to synthesize weather for
    data: pd.DataFrame
        Dataframe of formatted weather data to reference
    precip_dict: dict
        Dictionary of fit precipitation parameters 
    copulaetemp_dict: dict
        Dictionary of fit copula parameters to conditionally construct temperature 
    resolution: str
        The temporal resolution of the  data 
    validate: bool
        Flag for producing figures to validate each step of the generator
    dirpath: str
        Path for where to save the validation figures
    synthesize_kwargs: dict
        Keyword arguments related to the fit
    
    Returns
    -------
    synth_data: pd.DataFrame
        The synthesized weather data from the given observations and parameters
    """
    
    # synthesize kwargs    
    default_synthesize_kwargs = {"validation_samplesize_mult": 10,
                                 "figure_extension": "svg"}
    if not synthesize_kwargs: 
        synthesize_kwargs = default_synthesize_kwargs
    else:
        for k in default_synthesize_kwargs:
            if k not in synthesize_kwargs:
                synthesize_kwargs[k] = default_synthesize_kwargs[k]
    
    # validation
    global do_validation, validation_dirpath, validation_extension
    do_validation, validation_dirpath, validation_extension = validate, dirpath, synthesize_kwargs["figure_extension"]
     
    # remove all the rows with partially full or unfit years
    filtered_data = data[data["YEAR"].isin(precip_dict["log10_annual_precip"].index.values)]
    incomplete_years, full_years = [], []
    for year in sorted(set(filtered_data["YEAR"].values)):
        if len(set(filtered_data.loc[filtered_data["YEAR"] == year, "MONTH"].values)) == 12:
            full_years.append(int(year))
        else:
            incomplete_years.append(int(year))
    filtered_data = filtered_data[filtered_data["YEAR"].isin(full_years)]
    filtered_data.reset_index(drop=True, inplace=True)
    if resolution == "daily" and "DAY" in filtered_data.columns:
        filtered_data.astype({"SITE": str, "YEAR": int, "MONTH": int, "DAY": int, "PRECIP": float, "TEMP": float})
    else:
        filtered_data.astype({"SITE": str, "YEAR": int, "MONTH": int, "PRECIP": float, "TEMP": float})

    # collecting kNN years
    global kNN_dict
    kNN_dict = {"precip": [], "temp": {}}

    # synthesizing precipitation
    synth_precip = synthesize_precip(n, filtered_data, precip_dict, resolution, incomplete_years) 
    
    # conditionally synthesizing temperature from precipitation
    synth_pt = synthesize_pt_pairs(synth_precip, copulaetemp_dict, filtered_data, resolution) 

    # compare synth data to obs (if requested)
    if do_validation:
        print("Validating generated weather, this may take a moment...")
        add_samples = (synthesize_kwargs["validation_samplesize_mult"] - 1) * n
        if add_samples == 0:
            compare_synth_to_obs(validation_dirpath, validation_extension, synth_pt, filtered_data)
        else:
            kNN_dict = {"precip": [], "temp": {}}
            synth_precip_compare = synthesize_precip(add_samples, filtered_data, precip_dict, resolution, incomplete_years)
            synth_pt_compare = synthesize_pt_pairs(synth_precip_compare, copulaetemp_dict, filtered_data, resolution)
            synth_pt_compare["YEAR"] = synth_pt_compare["YEAR"].values + max(synth_pt["YEAR"].values)
            compare_pt = pd.concat([synth_pt, synth_pt_compare])
            if resolution == "daily" and "DAY" in compare_pt.columns:
                compare_pt.sort_values(by=["SITE", "YEAR", "MONTH", "DAY"], inplace=True)
            else:
                compare_pt.sort_values(by=["SITE", "YEAR", "MONTH"], inplace=True)
            compare_pt.reset_index(drop=True, inplace=True)
            compare_synth_to_obs(validation_dirpath, validation_extension, compare_pt, filtered_data)

    return synth_pt


def synthesize_precip(n_synth_years: int, data: pd.DataFrame, p_dict: dict, resolution: str, incomp_years: list[int]) -> np.array:
    """
    Manager function to synthesize precipitation

    Parameters
    ----------
    n_synth_years: int
        Number of years for the generator to synthesize weather for 
    data: pd.DataFrame
        Observed precipitation and temperature data
    p_dict: dict
        GMMHMM best-fitted model and corresponding parameters
    resolution: str
        Resolution of desired synthesized data
    incomp_years: list[int]
        List of years without the full set of months

    Returns
    -------
    synth_precip_monthly: pd.DataFrame
        Synthesized monthly precipitation
    """
    
    def precip_kNN_disaggregation(precip_log10annual_sample: np.array, 
                                  precip_obs: pd.DataFrame, 
                                  precip_log10annual_obs: pd.DataFrame) -> np.array:
        """
        Function to perform a k-NN disaggregation scheme for the log10(annual) precipitation 
        data specifically, adapted from ideas in Lall & Sharma (1996), 
        Apipattanavis et al. (2007), Nowak et al. (2010), and Quinn et al. (2020, supplmental). 
        This technique is non-parametric and therefore requires existing observations
        
        Parameters
        ----------
        precip_log10annual_sample: np.array
            Sample taken from the GMMHMM of the log10 annual precipitation data
        precip_obs: pd.DataFrame
            Observed precipitation data at the appropriate resolution
        precip_log10annual_obs: pd.DataFrame
            log10(annual)-transformed precipitation data as determined in
            the GMMHMM fit

        Returns
        -------
        disaggregated_sample: pd.DataFrame
            k-NN disaggregated monthly precipitation sample
        """

        # (0) convert to real-space from synth log-space
        n_months = 12
        synth_annual = 10.**precip_log10annual_sample

        # (1) create k, weights -- k recommended to be int(sqrt(n)), where n is the number of years
        k = round(np.sqrt(len(years)))
        w = np.array([(1 / j) for j in range(1, k+1)]) / sum([(1 / j) for j in range(1, k+1)])

        # (2) spatial averages for observations and sample
        obs_spatial_avg = np.nanmean(10.**precip_log10annual_obs.values, axis=1)
        synth_spatial_avg = np.nanmean(synth_annual, axis=1)

        # (3) choose one of the k closest observed years 
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            year_obs_pair = np.reshape([[years[i], obs_spatial_avg[i]] for i in range(len(years))], newshape=(len(years), 2))
        kNN_selected_years = np.full(shape=precip_log10annual_sample.shape[0], fill_value=np.nan)
        disaggregated_sample = np.full(shape=(precip_log10annual_sample.shape[0], n_months, len(sites)), fill_value=np.nan)
        for j, sum_synth_year in enumerate(synth_spatial_avg):
            # (4) calculate Manhattan distance (since 1D) between individual synthetic and all obs
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", DeprecationWarning)
                year_synth_dist = np.reshape([[year_obs_pair[i, 0], abs(sum_synth_year - year_obs_pair[i, 1])] for i in range(len(years))], newshape=(len(years), 2))
            sorted_year_dist = year_synth_dist[year_synth_dist[:, 1].argsort()]
            # (5) choose a year from the set using pre-determined weights
            kNN_selected_years[j] = rng.choice(sorted_year_dist[:k, 0], p=w)
        kNN_dict["precip"] = kNN_selected_years

        # (6) maintain temporal proportionality: synth_{month} = synth_{year} * hist_{month}/hist_{year}
        for i, kNN_selected_year in enumerate(kNN_selected_years):
            year_idx = precip_obs["YEAR"] == kNN_selected_year
            for s, site in enumerate(sites):
                site_idx = precip_obs["SITE"] == site
                if resolution == "monthly" or "DAY" not in precip_obs.columns:
                    kNN_selected_monthlies = precip_obs.loc[year_idx & site_idx, "PRECIP"].values
                else:
                    kNN_selected_monthlies = []
                    for m in range(1, n_months+1):
                        month_idx = precip_obs["MONTH"] == m
                        kNN_selected_monthlies.append(np.nansum(precip_obs.loc[month_idx & year_idx & site_idx, "PRECIP"].values))
                    kNN_selected_monthlies = np.array(kNN_selected_monthlies)
                disaggregated_sample[i, :, s] = synth_annual[i, s] * (kNN_selected_monthlies / sum(kNN_selected_monthlies))

        return disaggregated_sample
 
    rng = np.random.default_rng()
    filtered_annual_data = p_dict["log10_annual_precip"].drop(incomp_years)
    sites, years = sorted(set(data["SITE"].values)), sorted(set(filtered_annual_data.index.values))
    annual_sample = p_dict["model"].sample(n_synth_years)[0]
    precip_data = data[data.columns[:list(data.columns).index("PRECIP")+1]].copy()
    return precip_kNN_disaggregation(annual_sample, precip_data, filtered_annual_data) 


def synthesize_pt_pairs(synth_prcp: np.array, t_dict: dict, pt_df: pd.DataFrame, resolution: str) -> pd.DataFrame:
    """
    Manager function to conditionally synthesize temperature from 
    precipitation

    Parameters
    ----------
    synth_prcp: pd.DataFrame
        Synthesized precipitation at monthly resolution
    t_dict: dict
        Copula best-fitted, spatially-averaged models and corresponding parameters
    pt_df: pd.DataFrame
        Temporally-formatted observed precipitation and temperature data
    resolution: str
        Time resolution of the synthesized data

    Returns
    -------
    synth_monthly_df: pd.DataFrame
        Synthesized monthly precipitation and temperature
    """
    
    def conditionally_simulate_uT(poP: np.array, cop_list: list) -> np.array:
        """
        Conditionally simulate temperature pseudo-observations from
        synthesized precipitation pseudo-observations using the 
        appropriate copula family

        Parameters
        ----------
        poP: np.array
            Synthesized precipitation pseudo-observations
        cop_list: list
            List with the copula object and family name

        Returns
        -------
        poT: np.array
            Conditionally sampled temperature pseudo-observations
        """

        cop_obj, cop_name = cop_list[0], cop_list[1]
        if cop_name == "Independence":
            # v = d/du [C(u,v)] --> since C(u,v) = u*v, marginal v *is* the inverse of the conditional CDF
            poT = rng.random(size=len(poP))
        if cop_name == "Frank":
            # v = inverse of the conditional CDF -- c(v|u)^{-1} -- so the ppf of the copula given u
            y = rng.random(size=len(poP)) 
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=DeprecationWarning)
                try:
                    poT = cop_obj.percent_point(y, poP)
                except ValueError:
                    y = rng.random(size=len(poP))
                    poT = cop_obj.percent_point(y, poP)
        if cop_name == "Gaussian":
            # (1) conditional sampling starts with the Cholesky decomposition of the Gaussian parameter
            # (2) transform to normal distribution for poP, generate on normal distribution for y
            # (3) matrix multiply to get sample simulation given input
            # (4) temperature marginals are the cdf of the temperature half of the conditional sample
            A = np.tril(np.linalg.cholesky(cop_obj.correlation.values))
            normP = scipy.stats.norm.ppf(poP)
            y = scipy.stats.norm.ppf(rng.random(size=len(poP)))
            cond_samp = A @ np.array([normP, y])
            poT = scipy.stats.norm.cdf(cond_samp[1])

        return poT

    def temp_kNN_disaggregation(t_monthly_synth: np.array, pt_monthly_obs: pd.DataFrame, mnth: int) -> np.array:
        """
        Function to perform a k-NN disaggregation scheme for the spatially-averaged
        temperature data specifically, adapted from ideas in Lall & Sharma (1996), 
        Apipattanavis et al. (2007), Nowak et al. (2010), and Quinn et al. (2020, supplmental). 
        This technique is non-parametric and therefore requires existing observations
        
        Parameters
        ----------
        t_monthly_synth: np.array
            Synthesized spatially-averaged temperature at monthly resolution
        pt_monthly_obs: pd.DataFrame
            Observed precipitation and temperature at monthly resolution
        mnth: int
            Month of the year, as an integer value

        Returns
        -------
        disaggregated_sample: pd.DataFrame
            k-NN disaggregated monthly temperature sample
        """
        
        k = round(np.sqrt(len(years)))
        w = np.array([(1 / j) for j in range(1, k+1)]) / sum([(1 / j) for j in range(1, k+1)])
 
        # (1) spatial averages for obs and synth
        obs_month_idx = pt_monthly_obs["MONTH"] == mnth
        obs_spatial_avg = []
        for year in years:
            obs_year_idx = pt_monthly_obs["YEAR"] == year
            obs_temps = pt_monthly_obs.loc[obs_month_idx & obs_year_idx, "TEMP"].values
            obs_spatial_avg.append(np.inf if all(np.isnan(obs_temps)) else np.nanmean(obs_temps)) 
        obs_spatial_avg = np.array(obs_spatial_avg)
        synth_spatial_avg = t_monthly_synth 

        # (2) link the summed historic year with the year itself, empty vector to fill with year choices
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            year_obs_pair = np.reshape([[years[i], obs_spatial_avg[i]] for i in range(len(years))], newshape=(len(years), 2))
        kNN_selected_years = np.full(shape=len(synth_spatial_avg), fill_value=np.nan)
        for j, sa_synth_year in enumerate(synth_spatial_avg):
            # (3) calculate Manhattan distance (since 1D) between individual synthetic and all obs
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", DeprecationWarning)
                year_synth_dist = np.reshape([[year_obs_pair[i, 0], abs(sa_synth_year - year_obs_pair[i, 1])] for i in range(len(years))], newshape=(len(years), 2))
            sorted_year_dist = year_synth_dist[year_synth_dist[:, 1].argsort()]
            # (4) choose which year from the set of years using pre-determined weights
            kNN_selected_years[j] = rng.choice(sorted_year_dist[:k, 0], p=w)
        if mnth not in kNN_dict:
            kNN_dict["temp"][mnth] = kNN_selected_years

        # (5) construct a vector of spatial averages that match the selected kNN years, 
        # -- add noise to smooth out the emergent non-parametric banding
        kNN_spatial_avg = np.full(shape=len(synth_spatial_avg), fill_value=np.nan)
        for i, kNN_selected_year in enumerate(kNN_selected_years):
            knn_year_idx = pt_monthly_obs["YEAR"] == kNN_selected_year
            kNN_selected_station_values = pt_monthly_obs.loc[knn_year_idx & obs_month_idx, "TEMP"].values
            kNN_spatial_avg[i] = np.nanmean(kNN_selected_station_values)
        resids = synth_spatial_avg - kNN_spatial_avg
        disaggregated_sample = np.full(shape=(len(synth_spatial_avg), len(sites)), fill_value=np.nan)
        for i, kNN_selected_year in enumerate(kNN_selected_years):
            knn_year_idx = pt_monthly_obs["YEAR"] == kNN_selected_year
            kNN_selected_station_values = pt_monthly_obs.loc[knn_year_idx & obs_month_idx, "TEMP"].values 
            noise = rng.normal(loc=0., scale=np.nanstd(resids), size=1) 
            disaggregated_sample[i, :] = ((synth_spatial_avg[i] + 273.15) * 
                                          ((kNN_selected_station_values + 273.15) / (np.nanmean(kNN_selected_station_values) + 273.15 + noise))) - 273.15 

        return disaggregated_sample
    
    def daily_kNN_disaggregation(synth_month_df: pd.DataFrame, obs_daily_df: pd.DataFrame) -> pd.DataFrame:
        """
        Using previously discovered k-NN years and months, convert monthly
        WX data to daily data. This technique is non-parametric and therefore 
        requires existing observations
        
        Parameters
        ----------
        synth_month_df: pd.DataFrame
            Synthetic precipitation and temperature at monthly resolution
        obs_daily_df: pd.DataFrame
            Observed precipitation and temperature at daily resolution

        Returns
        -------
        disaggregated_sample: pd.DataFrame
            k-NN disaggregated daily synthetic data
        """
        
        # average precip values per doy from observations
        doy_dict = {}
        for site in sorted(set(obs_daily_df["SITE"].values)):
            doy_site_idx = obs_daily_df["SITE"] == site
            doy_site_entry = obs_daily_df.loc[doy_site_idx]
            if site not in doy_dict: doy_dict[site] = {}
            for i in range(doy_site_entry.shape[0]):
                row_entry = doy_site_entry.iloc[i]
                doy = int(dt.datetime.strptime("{}-{}-{}".format(int(row_entry["YEAR"]), str(row_entry["MONTH"]).zfill(2), str(row_entry["DAY"]).zfill(2)), "%Y-%m-%d").strftime("%j"))
                if doy not in doy_dict[site]:
                    doy_dict[site][doy] = [row_entry["PRECIP"]]
                else:
                    doy_dict[site][doy].append(row_entry["PRECIP"])
            for doy in doy_dict[site]:
                doy_dict[site][doy] = float(np.nanmean(doy_dict[site][doy]))

        daily_dict = {}
        for site in sites:
            monthly_site_idx = synth_month_df["SITE"] == site
            monthly_site_entry = synth_month_df.loc[monthly_site_idx] 
            daily_site_idx = obs_daily_df["SITE"] == site
            daily_site_entry = obs_daily_df.loc[daily_site_idx]
            for mnth in sorted(set(monthly_site_entry["MONTH"].values)):    
                monthly_month_idx = monthly_site_entry["MONTH"] == mnth
                monthly_month_entry = monthly_site_entry.loc[monthly_month_idx]
                daily_month_idx = daily_site_entry["MONTH"] == mnth
                daily_month_entry = daily_site_entry.loc[daily_month_idx]
                for y in range(len(set(synth_month_df["YEAR"].values))):
                    monthly_year_idx = monthly_month_entry["YEAR"] == y+1
                    monthly_year_entry = monthly_month_entry.loc[monthly_year_idx]
                    daily_precip_year_idx = daily_month_entry["YEAR"] == kNN_dict["precip"][y]
                    daily_precip_year_entry = daily_month_entry.loc[daily_precip_year_idx]
                    daily_temp_year_idx = daily_month_entry["YEAR"] == kNN_dict["temp"][mnth][y]
                    daily_temp_year_entry = daily_month_entry.loc[daily_temp_year_idx]
                    sync_num_days = min([daily_precip_year_entry.shape[0], daily_temp_year_entry.shape[0]]) 
                    if np.nansum(daily_precip_year_entry["PRECIP"].values[:sync_num_days]) == 0:
                        sync_doys = []
                        for d in range(sync_num_days):
                            sync_doys.append(int(dt.datetime.strptime("{}-{}-{}".format(int(kNN_dict["precip"][y]),str(mnth).zfill(2),str(d+1).zfill(2)), "%Y-%m-%d").strftime("%j")))
                        sync_precips = [doy_dict[site][sync_doy] for sync_doy in sync_doys]
                    else:
                        sync_precips = daily_precip_year_entry["PRECIP"].values[:sync_num_days]
                    sync_temps = daily_temp_year_entry["TEMP"].values[:sync_num_days]
                    for day in range(sync_num_days): 
                        prcp = sync_precips[day] * (monthly_year_entry["PRECIP"].values[0] / np.nansum(sync_precips))
                        temp = sync_temps[day] + (monthly_year_entry["TEMP"].values[0] - np.nanmean(sync_temps))
                        daily_dict[(site, y+1, mnth, day+1)] = [site, y+1, mnth, day+1, round(prcp, 4), round(temp, 2)] 
        daily_df = pd.DataFrame().from_dict(daily_dict, orient="index", columns=["SITE", "YEAR", "MONTH", "DAY", "PRECIP", "TEMP"])
        daily_df.sort_values(by=["SITE", "YEAR", "MONTH", "DAY"], inplace=True)
        daily_df.reset_index(drop=True, inplace=True)
        daily_df.astype({"SITE": str, "YEAR": int, "MONTH": int, "DAY": int, "PRECIP": float, "TEMP": float}) 
        return daily_df

    rng = np.random.default_rng()
    sites = sorted(set(pt_df["SITE"].values))
    years = [y for y in range(min(pt_df["YEAR"].values), max(pt_df["YEAR"].values)+1)]
    month_names, month_vals = list(t_dict.keys()), [m+1 for m in range(len(t_dict.keys()))]
    n_synth_years, n_months, n_sites = synth_prcp.shape
    synth_monthly_df = pd.DataFrame(columns=["SITE", "YEAR", "MONTH", "PRECIP", "TEMP"])
    synth_monthly_df["SITE"] = np.repeat(sites, n_synth_years * n_months)
    synth_monthly_df["YEAR"] = list(np.repeat([y+1 for y in range(n_synth_years)], n_months)) * n_sites
    synth_monthly_df["MONTH"] = month_vals * (n_synth_years * n_sites)

    if resolution == "daily" and "DAY" in pt_df.columns:
        pt_monthly_df_dict = {}
        for site in sorted(set(pt_df["SITE"].values)):
            site_idx = pt_df["SITE"] == site
            for year in sorted(set(pt_df["YEAR"].values)):
                year_idx = pt_df["YEAR"] == year
                for month in sorted(set(pt_df["MONTH"].values)):
                    month_idx = pt_df["MONTH"] == month
                    prcps = pt_df.loc[site_idx & year_idx & month_idx, "PRECIP"].values
                    prcp_sum = np.nan if len(prcps) == 0 else np.nansum(prcps)
                    temps = pt_df.loc[site_idx & year_idx & month_idx, "TEMP"].values
                    temp_avg = np.nan if len(temps) == 0 else np.nanmean(temps)
                    pt_monthly_df_dict[(site, year, month)] = [site, year, month, prcp_sum, temp_avg]
        pt_monthly_df = pd.DataFrame().from_dict(pt_monthly_df_dict, orient="index", columns=["SITE", "YEAR", "MONTH", "PRECIP", "TEMP"])
        pt_monthly_df.reset_index(drop=True, inplace=True)
        pt_monthly_df.astype({"SITE": str, "YEAR": int, "MONTH": int, "PRECIP": float, "TEMP": float})
    else:
        pt_monthly_df = pt_df
    
    for m, month in enumerate(month_names):
        month_idx = synth_monthly_df["MONTH"] == month_vals[m]
        
        # spatially average synth precip for the month
        sa_synth_prcp = synth_prcp[:, m, :].mean(axis=1)
        
        # transform the synthetic preciptation to residuals using the ARfit used in the copulas
        nP = len(sa_synth_prcp)
        ar1_synth_prcp_fit = t_dict[month]["PRECIP ARFit"].apply(sa_synth_prcp)
        full_ar1_prcp = np.array([np.nanmean(ar1_synth_prcp_fit.fittedvalues), *ar1_synth_prcp_fit.fittedvalues])
        resid_prcp = sa_synth_prcp - full_ar1_prcp

        # transform into uniform marginals
        uP = stats.rankdata(resid_prcp, method="average") / (nP+1)

        # conditional simulation of the uT | uP --> coming from {d/d(uP) [C(uP, uT)]}^{-1}
        uT = conditionally_simulate_uT(uP, t_dict[month]["BestCopula"])

        # transform from marginals to residuals (using CDF^{-1}) to data (using AR fit params)
        resid_temp = t_dict[month]["TEMP Resid Dist"].ppf(uT)
        obs_mean, obs_std = np.nanmean(t_dict[month]["TEMP"].astype(float)), np.nanstd(t_dict[month]["TEMP"].astype(float))
        sa_synth_temp_approx = (resid_temp + t_dict[month]["TEMP ARFit"].params[0]) / (1. - t_dict[month]["TEMP ARFit"].params[1]) 
        approx_mean, approx_std = np.nanmean(sa_synth_temp_approx), np.nanstd(sa_synth_temp_approx)
        sa_synth_temp = (sa_synth_temp_approx - approx_mean)*(obs_std/approx_std) + obs_mean

        # (parametric) conditional temperature can sample values WAY too high or low
        # -- if this happens, resample the conditional temperatures until it doesn't happen
        obs_temp = t_dict[month]["TEMP"].astype(float)
        obs_max_diff = np.abs(np.nanmax(obs_temp) - np.nanmin(obs_temp))
        while np.any(sa_synth_temp < np.nanmin(obs_temp) - obs_max_diff) or np.any(sa_synth_temp > np.nanmax(obs_temp) + obs_max_diff):
            uT = conditionally_simulate_uT(uP, t_dict[month]["BestCopula"])
            resid_temp = t_dict[month]["TEMP Resid Dist"].ppf(uT)
            sa_synth_temp_approx = (resid_temp + t_dict[month]["TEMP ARFit"].params[0]) / (1. - t_dict[month]["TEMP ARFit"].params[1]) 
            approx_mean, approx_std = np.nanmean(sa_synth_temp_approx), np.nanstd(sa_synth_temp_approx)
            sa_synth_temp = (sa_synth_temp_approx - approx_mean)*(obs_std/approx_std) + obs_mean

        # take the spatially-averaged temperatures and disaggregate to return per-station values
        synth_temp = temp_kNN_disaggregation(sa_synth_temp, pt_monthly_df, month_vals[m])
        
        # assign to dataframe 
        for s, site in enumerate(sites):
            site_idx = synth_monthly_df["SITE"] == site
            synth_monthly_df.loc[month_idx & site_idx, "PRECIP"] = synth_prcp[:, m, s]            
            synth_monthly_df.loc[month_idx & site_idx, "TEMP"] = synth_temp[:, s] 
    
    if resolution == "monthly" or (resolution == "daily" and "DAY" not in pt_df.columns):
        if resolution == "daily" and "DAY" not in pt_df.columns:
            warnings.warn("Input dataset at monthly resolution cannot be disaggregated to daily! Returning monthly...", UserWarning)
        return synth_monthly_df
    else:
        return daily_kNN_disaggregation(synth_monthly_df, pt_df) 




import numpy as np
import pandas as pd
import datetime as dt

import pytest
from swxg.model import SWXGModel
from swxg.test_wx import monthly, daily


@pytest.mark.filterwarnings("ignore::pytest.PytestUnraisableExceptionWarning")
def test_generator() -> None:  
    #model = SWXGModel(monthly) 
    model = SWXGModel(daily) 
    model.fit()
    synth_wx = model.synthesize()
    #print(synth_wx)



import numpy as np
import pandas as pd
import matplotlib as mpl
from matplotlib import pyplot as plt
import matplotlib as mpl
from statsmodels.graphics.tsaplots import plot_acf
import scipy
import math
from statsmodels.graphics.tsaplots import plot_acf, plot_pacf
from statsmodels.tsa.stattools import acf, pacf
from statsmodels.tsa.ar_model import AutoReg
import statsmodels.api as sm
import warnings
import datetime as dt


def squarest_subplots(n: int) -> tuple[int]:
    """
    For some input value, return the most rectangular (or square)
    2D arrangement of that input value

    Parameters
    ----------
    n: int
        Number of elements to rectangularize

    Returns
    -------
    rows: int
        Number of rows in the 2D arrangement
    cols: int
        Number of cols in the 2D arrangement
    """
    
    rows = int(n ** 0.5)
    cols = int(np.ceil(n / rows))
    return rows, cols


def validate_gmmhmm_states(dp: str, ext: str, min_states: int, max_states: int, lls: list[float], aics: list[float], bics: list[float]) -> None:
    """
    Validation figure for confirming the best-fitting number of states
    for the Gaussian mixture model hidden Markov model that represents
    precipitation

    Parameters
    ----------
    dp: str
        Filepath for saving the validation figure
    ext: str
        Filepath extension for saving the validation figure
    min_states: int
        The minimum number of attempted hidden states in fit
    max_states: int
        The maximum number of attempted hidden states in fit 
    lls: list[float]
        Log-likelihood calculation for each fit GMMHMM by number of states
    aics: list[float]
        AIC calculation for each fit GMMHMM by number of states
    bics: list[float]
        BIC calculation for each fit GMMHMM by number of states 
    """

    len_states = len(np.arange(min_states, max_states + 1))
    if len_states > len(lls):
        lls.extend([np.nan] * (len_states - len(lls)))
        aics.extend([np.nan] * (len_states - len(aics)))
        bics.extend([np.nan] * (len_states - len(bics)))
    num_states_fig, axis = plt.subplots()
    axis.grid() 
    axis.plot(np.arange(min_states, max_states + 1), aics, color="blue", marker="o", label="AIC")
    axis.plot(np.arange(min_states, max_states + 1), bics, color="green", marker="o", label="BIC")
    axis2 = axis.twinx()
    axis2.plot(np.arange(min_states, max_states + 1), lls, color="orange", marker="o", label="LL")
    axis.legend(handles=axis.lines + axis2.lines)
    axis.set_title("Validation of GMMHMM Best-Fitting Number of States")
    axis.set_xlabel("# States")
    axis.set_ylabel("Criterion Value [-, lower is better]")
    axis2.set_ylabel("Log-Likelihood [-, higher is better]")
    plt.tight_layout()
    num_states_fig.savefig("{}Validate_GMMHMM_NumStates.{}".format(dp, ext))
    plt.close()


def validate_explore_pt_dependence(dp: str, ext: str, pt_data: pd.DataFrame, good_years: list[int]) -> None:
    """
    Validation figure for exploring the Kendall and Spearman correlation 
    coefficients between precipitation and temperature. Significant
    positive or negative correlations implies a need for a copula to
    represent the conditional relationship between them.

    Parameters
    ----------
    dp: str
        Filepath for saving the validation figure
    ext: str
        Filepath extension for saving the validation figure
    pt_data: pd.DataFrame
        The precipitation and temperature data, as a DataFrame
    good_years: list[int]
        The list of years fit by the GMMHMM
    """
    
    sites = sorted(set(pt_data["SITE"].values))
    months = ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]

    # spatially-averaged correlation between precipitation and temperature
    spatial_corr_df = pd.DataFrame({"Kendall": np.nan, "Spearman": np.nan}, index=months)
    spatial_data_dict = {month: {"PRECIP": [], "TEMP": []} for month in months}
    for month in months:
        month_index = pt_data["MONTH"] == month
        for year in good_years:
            year_index = pt_data["YEAR"] == year
            ps = pt_data.loc[month_index & year_index, "PRECIP"].values
            p = np.nan if np.all(np.isnan(ps)) else np.nanmean(ps)
            ts = pt_data.loc[month_index & year_index, "TEMP"].values
            t = np.nan if np.all(np.isnan(ts)) else np.nanmean(ts)
            spatial_data_dict[month]["PRECIP"].append(p)
            spatial_data_dict[month]["TEMP"].append(t)
        spatial_corr_df.at[month, "Kendall"] = pd.DataFrame({"PRECIP": spatial_data_dict[month]["PRECIP"], 
                                                             "TEMP": spatial_data_dict[month]["TEMP"]}).corr(method="kendall")["PRECIP"]["TEMP"]
        spatial_corr_df.at[month, "Spearman"] = pd.DataFrame({"PRECIP": spatial_data_dict[month]["PRECIP"], 
                                                              "TEMP": spatial_data_dict[month]["TEMP"]}).corr(method="spearman")["PRECIP"]["TEMP"]
    
    # kendall/spearman correlation metric plot
    sa_corr_fig, axis = plt.subplots(nrows=1, ncols=1, figsize=(16, 9))
    sa_corr_fig.suptitle("Correlation of Spatially-Averaged Precip/Temp Data by Month")
    sa_corr_fig.supxlabel("Month"), sa_corr_fig.supylabel("Correlation Coefficient [-]")
    axis.grid()
    axis.set_ylim(-1, 1)
    axis.set_xticks(range(len(months)))
    axis.set_xticklabels(months, rotation=45)
    axis.hlines(0, xmin=0, xmax=11, colors="black", linestyles="dashed")
    axis.plot(range(len(months)), spatial_corr_df["Kendall"], marker="o", label=r"Kendall $\tau$")
    axis.plot(range(len(months)), spatial_corr_df["Spearman"], marker="o", label=r"Spearman $\rho$")
    axis.legend()
    plt.tight_layout()
    sa_corr_fig.savefig("{}Validate_Copulae_ExplorePTCorrelation_MonthlySpatialAverage.{}".format(dp, ext))
    plt.close()
     
    # plot scatterplot of spatially averaged precipitation and temperature
    pt_dist_fig = plt.figure(figsize=(14, 9))
    pt_dist_fig.suptitle("Direct Comparison of Spatially-Averaged Precipitation and Temperature | PRECIP on x-axis, TEMP on y-axis")
    sub_figs = pt_dist_fig.subfigures(3, 4)
    for i, sub_fig in enumerate(sub_figs.flat):
        axes = sub_fig.subplots(2, 2, gridspec_kw={"width_ratios": [4, 1], "height_ratios": [1, 3]})
        sub_fig.subplots_adjust(wspace=0, hspace=0)
        month = months[i]
        for j, axis in enumerate(axes.flat):
            if j == 0:
                axis.hist(spatial_data_dict[month]["PRECIP"], density=True, color="black")
                axis.set(xticks=[], yticks=[])
            if j == 1:
                axis.axis("off")
                axis.text(0.5, 0.5, month, transform=axis.transAxes, va="center", ha="center")
            if j == 2:
                axis.scatter(spatial_data_dict[month]["PRECIP"], spatial_data_dict[month]["TEMP"], marker="o", facecolors="none", edgecolors="black")
            if j == 3:
                axis.hist(spatial_data_dict[month]["TEMP"], density=True, color="black", orientation="horizontal")
                axis.set(xticks=[], yticks=[])
    pt_dist_fig.savefig("{}Validate_Copulae_PTDistribution_MonthlySpatialAverage.{}".format(dp, ext))
    plt.close()


def validate_pt_acf(dp: str, ext: str, pt_dict: dict, lag: int) -> None:
    """
    Validation figure for ACF fits on precipitation and temperature,
    as small multiples by month

    Parameters
    ----------
    dp: str
        Filepath for saving the validation figure
    ext: str
        Filepath extension for saving the validation figure
    pt_dict: pd.DataFrame
        The precipitation and temperature data organized by month, as a dict
    lag: int
        The lag considered in the autocorrelation function
    """
    
    for weather_var in ["PRECIP", "TEMP"]:
        weather_color = "royalblue" if weather_var == "PRECIP" else "firebrick"
        acf_fig, axes = plt.subplots(nrows=3, ncols=4, figsize=(14, 9), sharex="all", sharey="all")
        acf_fig.suptitle("{} ACF from AR({}) | Color=ACF from Raw Data, Black=ACF from Residuals | Black should be fully in band".format(weather_var.capitalize(), lag))
        acf_fig.supxlabel("Lag [-]"), acf_fig.supylabel("ACF [-]")
        months = list(pt_dict.keys())
        for i, axis in enumerate(axes.flat):
            axis.grid()
            plot_acf(ax=axis, x=pt_dict[months[i]][weather_var], color=weather_color, vlines_kwargs={"color": weather_color, "label": None})
            plot_acf(ax=axis, x=pt_dict[months[i]][weather_var + " ARFit"].resid, color="black", vlines_kwargs={"color": "grey", "label": None})
            axis.set(title=months[i])
        plt.tight_layout()
        acf_fig.savefig("{}Validate_Copulae_{}_ACF.{}".format(dp, weather_var.capitalize(), ext))
        plt.close()


def validate_pt_stationarity(dp: str, ext: str, pt_dict: dict, groups: int) -> None:
    """
    Validation figure for checking the stationarity of the precipitation
    and temperature residuals through the Mann-Whitney U test

    Parameters
    ----------
    dp: str
        Filepath for saving the validation figure
    ext: str
        Filepath extension for saving the validation figure
    pt_dict: pd.DataFrame
        The precipitation and temperature data organized by month, as a dict
    groups: int
        Number of groups to consider for the stationarity test
    """
    
    groups = 2 if groups < 2 else groups
    bar_width = 1. / groups
    for weather_var in ["PRECIP", "TEMP"]:
        stationarity_fig, axis = plt.subplots(nrows=1, ncols=1, figsize=(16, 9))
        stationarity_fig.suptitle("{} Residuals Stationarity Check with {} Groups | Above Dashed Line Implies Stationarity".format(weather_var.capitalize(), groups))
        stationarity_fig.supxlabel("Month"), stationarity_fig.supylabel("Mann-Whitney U p-Value [-]")
        months = list(pt_dict.keys())
        mwu_pvalues = np.full(shape=(len(months), groups-1), fill_value=np.nan)
        for m, month in enumerate(months):
            resids = pt_dict[month][weather_var + " ARFit"].resid
            group_chunk = len(resids)//groups if len(resids) % groups == 0 else len(resids)//groups+1
            group_data = []
            for n in range(groups):
                group_data.append(resids[n*group_chunk:(n+1)*group_chunk])
            for g in range(groups-1):
                x_data, y_data = np.array(group_data[g], dtype=float), np.array(group_data[g+1], dtype=float)
                x_mask, y_mask = ~np.isnan(x_data), ~np.isnan(y_data)
                mwu_pvalues[m, g] = scipy.stats.mannwhitneyu(x=x_data[x_mask], y=y_data[y_mask])[1]

        # actually plotting
        axis.grid()
        axis.set(ylim=[0, 1])
        for g in range(groups-1):
            axis.bar([x+g*bar_width for x in range(len(months))], mwu_pvalues[:, g], width=bar_width, zorder=10)
        axis.hlines(0.05, -1, 12, color="black", linestyles="dashed", zorder=11)
        axis.set_xticks([m+(groups-2)*(bar_width/2) for m in range(len(months))])
        axis.set_xticklabels(labels=months, rotation=45)
        plt.tight_layout()
        stationarity_fig.savefig("{}Validate_Copulae_{}_ResidStationarity{}Groups.{}".format(dp, weather_var.title(), groups, ext))
        plt.close()


def validate_pt_dependence_structure(dp: str, ext: str, pt_dict: dict) -> None:
    """
    Validation figure for checking the dependence structure of 
    copula families for the precipitation and temperature residuals 
    through K-plots
    (Genest & Boies, 2003: https://www.jstor.org/stable/30037296) 

    Parameters
    ----------
    dp: str
        Filepath for saving the validation figure
    ext: str
        Filepath extension for saving the validation figure
    pt_dict: dict
        The precipitation and temperature data organized by month, as a dict
    """
    
    # define the functional form of the integrand for W_i:n
    def W_inIntegrand(w, idx, num):
        scale = num * math.comb(num - 1, idx - 1)
        u = w - (w * np.log(w))
        return scale * w * u ** (idx - 1) * (1 - u) ** (num - idx) * -np.log(w)

    kplots_fig, axes = plt.subplots(nrows=3, ncols=4, figsize=(16, 9), sharex="all", sharey="all")
    kplots_fig.suptitle("Monthly K-Plots | Above Dashed Line: (+) Dependence, Below Dashed Line: (-) Dependence ")
    kplots_fig.supxlabel("$W_{i:n}$"), kplots_fig.supylabel("$H_{(i)}$")
    months = list(pt_dict.keys())
    for i, axis in enumerate(axes.flat):
        n = len(pt_dict[months[i]]["PRECIP ARFit"].resid)
        H_i, W_in = np.array([]), np.array([])

        # calculate W_in, H_i
        for ii in range(n):
            p_i = pt_dict[months[i]]["PRECIP ARFit"].resid[ii]
            T_i = pt_dict[months[i]]["TEMP ARFit"].resid[ii]
            if np.isnan(p_i) or np.isnan(T_i):
                continue
            H = (1 / (n - 1)) * sum([1 for j in range(n) if j != ii and
                                     pt_dict[months[i]]["PRECIP ARFit"].resid[j] <= p_i and
                                     pt_dict[months[i]]["TEMP ARFit"].resid[j] <= T_i])
            H_i = np.append(H_i, H)
            W_in = np.append(W_in, scipy.integrate.quad(W_inIntegrand, a=0, b=1, args=(ii + 1, n))[0])
        H_i.sort()

        # plot H_(i) against W_in
        axis.grid()
        axis.set_title(months[i])
        axis.plot([jj / n for jj in range(n)], [jj / n for jj in range(n)], c="black", linestyle="dashed")
        axis.plot(W_in, H_i, c="magenta")
    plt.tight_layout()
    kplots_fig.savefig("{}Validate_Copulae_KPlots.{}".format(dp, ext))
    plt.close()


def validate_pt_fits(dp: str, ext: str, data_df: pd.DataFrame, precip_dict: dict, temp_dict: dict, val_figs: list[str]) -> None:
    """
    Validation manager for all of the figures that can be generated
    after fitting precipitation and temperature

    Parameters
    ----------
    dp: str
        Filepath for saving the validation figure
    ext: str
        Filepath extension for saving the validation figure
    data_df: pd.DataFrame
        Temporally-formated precipitation and temperature data, as a dataframe
    precip_dict: dict 
        The fitted precipitation data, as a dict
    temp_dict: dict
        The fitted temperature and copulae data, as a dict
    val_figs: list[str]
        Which of the validation figure methods to use/plot, can be 'precip' or 'copula'
    """
     
    # functions to call
    #validate_obs_spatial_temporal_correlations(dp, ext, data_df, precip_dict, temp_dict, val_figs)
    if "precip" in val_figs:
        validate_gmmhmm_statistics(dp, ext, data_df, precip_dict)
    if "copula" in val_figs:
        validate_copulae_statistics(dp, ext, data_df, temp_dict)


def validate_obs_spatial_temporal_correlations(dp: str, ext: str, data: pd.DataFrame, p_dict: dict, t_dict: dict, val_wvar: list[str]) -> None:
    """
    Validation figures for all the observed precipitation and temperature 
    spatial correlations, using the Pearson method (since there's no 
    comparison between parameters); precipitation temporal (Markovian) 
    structure at annual and monthly levels. CURRENTLY NOT PRODUCED
    since ``compare_synth_to_obs()`` replicates these.

    Parameters
    ----------
    dp: str
        Filepath for saving the validation figure
    ext: str
        Filepath extension for saving the validation figure
    data: pd.DataFrame
        Temporally-formated precipitation and temperature data, as a dataframe
    p_dict: dict 
        The fitted precipitation data, as a dict
    t_dict: dict
        The fitted temperature and copulae data, as a dict
    val_wvar: list[str]
        Produce just precipitation, temperature, or both
    """
    
    sites = sorted(set(data["SITE"].values))
    good_years = list(p_dict["log10_annual_precip"].index)
    years = [y for y in range(min(good_years), max(good_years)+1)]
    month_names = list(t_dict.keys())
    month_names_to_nums_dict = {"Jan": 1, "Feb": 2, "Mar": 3, "Apr": 4, "May": 5, "Jun": 6,
                                "Jul": 7, "Aug": 8, "Sep": 9, "Oct": 10, "Nov": 11, "Dec": 12}
    empty_spatial_df = pd.DataFrame(index=years, columns=sites, dtype=float)
    spatial_dict = {"p": {k: empty_spatial_df.copy() for k in ["annual", *month_names]}, 
                    "T": {k: empty_spatial_df.copy() for k in ["annual", *month_names]}}
    temporal_dict = {"annual": {}, "monthly": {}}
    for site in sites:
        site_idx = data["SITE"] == site
        site_entry = data.loc[site_idx]
        for year in years:
            year_idx = site_entry["YEAR"] == year
            year_entry = site_entry.loc[year_idx]
            precip_annual = np.nan if len(year_entry["PRECIP"].values) == 0 else np.nansum(year_entry["PRECIP"].values)
            temp_annual = np.nan if len(year_entry["TEMP"].values) == 0 else np.nanmean(year_entry["TEMP"].values)
            # -- spatial
            spatial_dict["p"]["annual"].at[year, site] = precip_annual 
            spatial_dict["T"]["annual"].at[year, site] = temp_annual
            # -- temporal
            temporal_dict["annual"][(site, year)] = [site, year, np.log10(precip_annual)]
            for month in month_names:
                month_num = month_names_to_nums_dict[month]
                month_idx = year_entry["MONTH"] == month_num
                month_entry = year_entry.loc[month_idx]
                precip_monthly = np.nan if len(month_entry["PRECIP"].values) == 0 else np.nansum(month_entry["PRECIP"].values)
                temp_monthly = np.nan if len(month_entry["TEMP"].values) == 0 else np.nanmean(month_entry["TEMP"].values)
                # -- spatial
                spatial_dict["p"][month].at[year, site] = precip_monthly
                spatial_dict["T"][month].at[year, site] = temp_monthly
                #  -- temporal
                temporal_dict["monthly"][(site, year, month)] = [site, year, month, precip_monthly]
    temporal_dict["annual"] = pd.DataFrame().from_dict(temporal_dict["annual"], orient="index", columns=["SITE", "YEAR", "PRECIP"])
    temporal_dict["annual"].reset_index(drop=True, inplace=True)
    temporal_dict["annual"].astype({"SITE": str, "YEAR": int, "PRECIP": float})
    temporal_dict["monthly"] = pd.DataFrame().from_dict(temporal_dict["monthly"], orient="index", columns=["SITE", "YEAR", "MONTH", "PRECIP"])
    temporal_dict["monthly"].reset_index(drop=True, inplace=True)
    temporal_dict["monthly"].astype({"SITE": str, "YEAR": int, "MONTH": str, "PRECIP": float})

    # plot spatial
    corr_cmap = plt.get_cmap("gnuplot", 21)
    if ("precip" in val_wvar) and ("copula" in val_wvar): 
        wvars = ["p", "T"]
    else:
        wvars = ["p"] if "precip" in val_wvar else ["T"]
    for k in wvars:
        wvar = "precip" if k == "p" else "temp"
        wvar_dict = spatial_dict[k]
        for kk in wvar_dict.keys():
            corr_df = wvar_dict[kk].corr(method="pearson")
            spatial_fig, axis = plt.subplots(nrows=1, ncols=1, figsize=(9, 9))
            corr_colors = axis.imshow(corr_df.values, vmin=0, vmax=1, cmap=corr_cmap)
            spatial_fig.colorbar(corr_colors, label="{} {} Spatial Correlation [Pearson, -]".format(kk.title(), wvar.title()))
            plt.xticks(range(len(sites)), sites, rotation=75)
            plt.yticks(range(len(sites)), sites, rotation=0)
            spatial_fig.savefig("{}Validate_SpatialCorrelation_{}_{}.{}".format(dp, kk.title(), wvar.title(), ext))
            plt.close()

    # plot temporal
    rs, cs = squarest_subplots(len(sites)) 
    acf_handle = mpl.lines.Line2D([], [], color="royalblue", marker="o")
    racf_handle = mpl.lines.Line2D([], [], color="rebeccapurple", marker="o")
    pacf_handle = mpl.lines.Line2D([], [], color="chocolate", marker="o")
    for time_scale in ["annual", "monthly"]:
        temporal_fig, axes = plt.subplots(nrows=rs, ncols=cs, figsize=(16, 9), sharex="all", sharey="all")
        temporal_fig.suptitle("ACF and PACF for {} Precip".format(time_scale.title()))
        temporal_fig.supxlabel("Lag"), temporal_fig.supylabel("Index Value [-]")
        for a, axis in enumerate(axes.flat):
            axis.grid()
            site = sites[a]
            site_idx = temporal_dict[time_scale]["SITE"] == site
            site_entry = temporal_dict[time_scale].loc[site_idx]
            precip_data = site_entry["PRECIP"].values
            precip_data = precip_data[~np.isnan(precip_data)]
            raw_acf = plot_acf(ax=axis, x=precip_data, use_vlines=False, color="royalblue")
            raw_ar1fit = AutoReg(precip_data, lags=[1]).fit()
            resid_acf = plot_acf(ax=axis, x=raw_ar1fit.resid, use_vlines=False, color="rebeccapurple")
            raw_pacf = plot_pacf(ax=axis, x=precip_data, method="ywm", use_vlines=False, color="chocolate")
            if a == 0: axis.legend(handles=[acf_handle, racf_handle, pacf_handle], labels=["Obs ACF", "Obs AR(1) Resid ACF", "Obs PACF"])
            axis.set(title=site, xlim=(-0.25, 11.25))
            axis.hlines(0, xmin=0, xmax=11, linestyles="dashed", color="black")
        plt.tight_layout()
        temporal_fig.savefig("{}Validate_GMMHMM_MarkovianStructure_{}.{}".format(dp, time_scale.title(), ext))
        plt.close()


def validate_gmmhmm_statistics(dp: str, ext: str, data: pd.DataFrame, p_dict: dict) -> None:
    """
    Validation figures for the precipitation GMMHMM, confirming:
    
    * that the transition between hidden states is Markovian (if 
      more than one state is found)
    * the solved hidden state as a function of date (year)
    * the transition probabilities between states are sensible
    * that the transformed precipitation data is Gaussian
    
    Parameters
    ----------
    dp: str
        Filepath for saving the validation figure
    ext: str
        Filepath extension for saving the validation figure
    data: pd.DataFrame
        Temporally-formated precipitation and temperature data, as a dataframe
    p_dict: dict
        The fitted precipitation data, as a dict
    """

    sites = sorted(set(data["SITE"].values))
    good_years = list(p_dict["log10_annual_precip"].index)
    years = [y for y in range(min(good_years), max(good_years)+1)]
    rs, cs = squarest_subplots(len(sites)) 
    state_colors = ["blue", "green", "cyan", "magenta", "yellow"]

    # Q-Q plots -- are the log10-transformed annual precipitation data normal (they should be) 
    # --> confirming "GMM" in GMMHMM 
    qq_fig, axes = plt.subplots(nrows=p_dict["num_gmmhmm_states"], ncols=len(sites), figsize=(16, 9))
    qq_fig.suptitle("Q-Q Plot of States' Log-Normal Distributions vs. Annual Data")
    qq_fig.supxlabel("Theoretical Quantiles [-]"), qq_fig.supylabel("Data Quantiles [-]")
    for st in range(p_dict["num_gmmhmm_states"]):
        for si in range(len(sites)): 
            axis = axes[st, si] if p_dict["num_gmmhmm_states"] > 1 else axes if len(sites) == 1 else axes[si]
            axis.grid()
            if st == 0: axis.set(title=sites[si])
            mu, std = p_dict["means"][st][si], p_dict["stds"][st][si]
            state_leg_str = ["State {}".format(st)]
            state_idx = p_dict["hidden_states"] == st
            data = p_dict["log10_annual_precip"][sites[si]].values[state_idx]
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=UserWarning)
                sm.qqplot(ax=axis, data=data, dist=scipy.stats.norm, loc=mu, scale=std, line="45", 
                          **{"markerfacecolor": mpl.colors.to_rgba(state_colors[st], 0.67), 
                             "markeredgecolor": mpl.colors.to_rgba(state_colors[st], 0.67)})
            if si == 0: 
                axis.legend(state_leg_str)
            axis.set_xlabel(""), axis.set_ylabel("")
    plt.tight_layout()
    qq_fig.savefig("{}Validate_GMMHMM_QQs.{}".format(dp, ext))
    plt.close()
    
    # ACF and PACF plots -- are the hidden states Markovian (they should be)
    # --> confirming "HMM" in GMMHMM  
    if p_dict["num_gmmhmm_states"] > 1:
        hiddenstate_markov_fig, axes = plt.subplots(nrows=2, ncols=1, figsize=(16, 9), sharex="all")
        hiddenstate_markov_fig.suptitle("Markovian Structure of GMMHMM Hidden States")
        hiddenstate_markov_fig.supxlabel("Lag")
        for i, axis in enumerate(axes.flat):
            axis.grid()
            if i == 0:
                plot_acf(ax=axis, x=p_dict["hidden_states"], vlines_kwargs={"color": "black"})
                axis.set(ylabel="ACF [-]", xlim=(-0.25, 16.25))
                axis.hlines(0, xmin=0, xmax=24, color="red")
            else:
                plot_pacf(ax=axis, x=p_dict["hidden_states"], method="ywm", vlines_kwargs={"color": "black"})
                axis.set(ylabel="PACF [-]", xlim=(-0.25, 16.25))
                axis.hlines(0, xmin=0, xmax=24, color="red")
        plt.tight_layout()
        hiddenstate_markov_fig.savefig("{}Validate_GMMHMM_HiddenStateMarkovStructure.{}".format(dp, ext))
        plt.close()

    # plot the transition probability matrix 
    tprob_fig, axes = plt.subplots(nrows=1, ncols=1, figsize=(9, 9), sharex="all", sharey="all")
    tprob_fig.suptitle("Transition Probabilities from GMMHMM")
    tprob_im = axes.matshow(p_dict["t_probs"], vmin=0, vmax=1, cmap="gnuplot")
    for i in range(p_dict["t_probs"].shape[0]):
        for j in range(p_dict["t_probs"].shape[1]):
            axes.text(j, i, '{:.3f}'.format(p_dict["t_probs"][i, j]), ha='center', va='center', backgroundcolor=[0, 0, 0, 0.25], color="white")
    axes.set(xticks=range(p_dict["num_gmmhmm_states"]), xticklabels=["To State {}".format(s) for s in range(p_dict["num_gmmhmm_states"])],
             yticks=range(p_dict["num_gmmhmm_states"]), yticklabels=["From State {}".format(s) for s in range(p_dict["num_gmmhmm_states"])])
    tprob_fig.colorbar(tprob_im, label="Probability [-]")
    plt.tight_layout()
    tprob_fig.savefig("{}Validate_GMMHMM_TransitionProbabilities.{}".format(dp, ext))
    plt.close()


def validate_copulae_statistics(dp: str, ext: str, data: pd.DataFrame, t_dict: dict) -> None:
    """
    Validation figures for the copulae, confirming:
    * best-fitting copulae via AIC comparison
    * copulae plots vs. empirical copula
    
    Parameters
    ----------
    dp: str
        Filepath for saving the validation figure
    ext: str
        Filepath extension for saving the validation figure
    data: pd.DataFrame
        Temporally-formated precipitation and temperature data, as a dataframe
    t_dict: dict 
        The fitted precip/copulae/temperature data, as a dict
    """
    
    # aics
    metrics_fig = plt.figure(figsize=(16, 9))
    metrics_fig.suptitle('Monthly Copula Fitness Metrics | Validate against K-Plots and Empirical Copula')
    months, metrics = list(t_dict.keys()), ["AIC", "S_n", "T_n"]
    for a in range(len(metrics)):
        axis = metrics_fig.add_subplot(1, len(metrics), a+1, projection="polar")
        axis.spines["polar"].set_visible(False)
        metric = metrics[a]
        if metric == "AIC": axis.set_title("AIC")
        if metric == "S_n": axis.set_title("Cram\u00e9r von Mises")
        if metric == "T_n": axis.set_title("Kolmogorov-Smirnov")
        ind_metric, frk_metric, gau_metric = [np.nan] * len(months), [np.nan] * len(months), [np.nan] * len(months)
        for j, month in enumerate(months):
            fit_info = t_dict[month]["CopulaDF"]
            for idx in fit_info.index:
                if idx == "Independence": ind_metric[j] = fit_info.at[idx, metric]
                if idx == "Frank": frk_metric[j] = fit_info.at[idx, metric]
                if idx == "Gaussian": gau_metric[j] = fit_info.at[idx, metric]
        ind_metric = [*ind_metric, ind_metric[0]]
        frk_metric = [*frk_metric, frk_metric[0]]
        gau_metric = [*gau_metric, gau_metric[0]]
        radial_months = [*months, months[0]]
        month_label_loc = np.linspace(0, 2 * np.pi, num=len(radial_months))
        if not np.all(np.isnan(ind_metric)): axis.plot(month_label_loc, ind_metric, label="Ind.", c="grey")
        if not np.all(np.isnan(frk_metric)): axis.plot(month_label_loc, frk_metric, label="Frk.", c="blue")
        if not np.all(np.isnan(gau_metric)): axis.plot(month_label_loc, gau_metric, label="Gau.", c="red")
        axis.set_thetagrids(np.degrees(month_label_loc), labels=radial_months)
        axis.legend(loc="upper right")
    plt.tight_layout()
    metrics_fig.savefig("{}Validate_Copulae_FitMetrics.{}".format(dp, ext))
    plt.close()
 
    # copula comparison
    def calculate_empirical_copula(pseudo_observations, resolution=50):
        n_data, n_dim = pseudo_observations.shape[0], pseudo_observations.shape[1]
        u0 = np.linspace(0, 1, resolution)
        Cn = np.full(shape=(resolution,)*n_dim, fill_value=0.)
        for k in range(n_data):
            U1, U2 = pseudo_observations[k, 0], pseudo_observations[k, 1]
            for ii in range(resolution):
                for jj in range(resolution):
                    if U1 <= u0[ii] and U2 <= u0[jj]:
                        Cn[ii, jj] += 1.
        Cn /= n_data
        return Cn, u0
    def approximate_theoretical_copula(copula, marginals, use_df=False):
        C_theory = np.full(shape=(len(marginals), len(marginals)), fill_value=np.nan)
        for ii, u1 in enumerate(marginals):
            for jj, u2 in enumerate(marginals):
                data_df = pd.DataFrame(data={"uP": [u1], "uT": [u2]}, dtype=float)
                C_val = copula.cdf(data_df.values) if use_df else copula.cdf(data_df.values)
                C_theory[ii, jj] = C_val[0] if type(C_val) in [list, np.ndarray] else C_val
        return C_theory 
    copulae_fig, axes = plt.subplots(nrows=3, ncols=4, figsize=(16, 9), sharex="all", sharey="all")
    copulae_fig.suptitle("Copulae Comparison")
    copulae_fig.supxlabel("$U_P$"), copulae_fig.supylabel("$U_T$")
    legend_symbols, legend_labels = [mpl.patches.Patch(color="black")], ["Emp."]
    for i, axis in enumerate(axes.flat):
        month = months[i]
        p_obs = np.array([t_dict[month]["PRECIP pObs"], t_dict[month]["TEMP pObs"]]).T
        emp_C, u = calculate_empirical_copula(p_obs)
        axis.grid()
        axis.set_title(month)
        contour_levels = [(i + 1) / 10 for i in range(10)]
        eContour = axis.contour(u, u, emp_C, levels=contour_levels, colors="black", vmin=0, vmax=1, linewidths=0.75)
        axis.clabel(eContour, inline=True, fontsize=10)
        copula_fit_df = t_dict[month]["CopulaDF"]
        for family in copula_fit_df.index.values:
            if family == "Independence":
                axis.contour(u, u, approximate_theoretical_copula(copula_fit_df.at[family, "Copula"], u),
                             levels=contour_levels, colors="grey", vmin=0, vmax=1, linewidths=0.75, linestyles="dashed")
                if i == 8: legend_symbols.append(mpl.patches.Patch(color="grey")), legend_labels.append("Ind.")
            if family == "Frank":
                axis.contour(u, u, approximate_theoretical_copula(copula_fit_df.at[family, "Copula"], u),
                             levels=contour_levels, colors="blue", vmin=0, vmax=1, linewidths=0.75, linestyles="dashed")
                if i == 8: legend_symbols.append(mpl.patches.Patch(color="blue")), legend_labels.append("Frk.")
            if family == "Gaussian":
                axis.contour(u, u, approximate_theoretical_copula(copula_fit_df.at[family, "Copula"], u, use_df=True),
                             levels=contour_levels, colors="red", vmin=0, vmax=1, linewidths=0.75, linestyles="dashed")
                if i == 8: legend_symbols.append(mpl.patches.Patch(color="red")), legend_labels.append("Gau.")
        if i == 8: axis.legend(handles=legend_symbols, labels=legend_labels, loc="lower left")
        axis.scatter(p_obs[:, 0], p_obs[:, 1], c="black", s=1)
    plt.tight_layout()
    copulae_fig.savefig("{}Validate_Copulae_Comparison.{}".format(dp, ext))
    plt.close()


def compare_synth_to_obs(dp: str, ext: str, synth_df: pd.DataFrame, obs_df: pd.DataFrame) -> None:
    """
    Comparing the synthesized WX data with the observed WX data through
    a variety of statistical and visual tests

    Parameters
    ----------
    dp: str
        Filepath for saving the validation figure
    ext: str
        Filepath extension for saving the validation figure
    synth_df: pd.DataFrame
        Synthesized precipitation and temperature data at defined resolution
    obs_df: pd.DataFrame 
        Observed precipitation and temperature data at defined resolution
    """
    
    sites = sorted(set(obs_df["SITE"].values))
    month_names = ["Jan", "Feb", "Mar", "Apr", "May", "Jun",
                   "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]
    resolution = "monthly" if "DAY" not in obs_df.columns and "DAY" not in synth_df.columns else "daily"

    # annual precip histograms
    r_sites, c_sites = squarest_subplots(len(set(obs_df["SITE"].values)))
    comp_gmmhmm_fig, axes = plt.subplots(nrows=r_sites, ncols=c_sites, figsize=(16, 9), sharex="all", sharey="all")
    comp_gmmhmm_fig.suptitle("Comparison of Obs and Synth Annual Precipitation via Histogram")
    comp_gmmhmm_fig.supxlabel("Total Annual Precipitation [m]"), comp_gmmhmm_fig.supylabel("Probability Density [-]")
    for a, axis in enumerate(axes.flat):
        if a >= len(sites): continue
        obs_site_idx, synth_site_idx = obs_df["SITE"] == sites[a], synth_df["SITE"] == sites[a]
        obs_site_entry, synth_site_entry = obs_df.loc[obs_site_idx], synth_df.loc[synth_site_idx]
        obs_annuals, synth_annuals = [], []
        for year in sorted(set(obs_site_entry["YEAR"].values)):
            obs_annuals.append(np.nansum(obs_site_entry.loc[obs_site_entry["YEAR"] == year, "PRECIP"].values))
        for year in sorted(set(synth_site_entry["YEAR"].values)):
            synth_annuals.append(np.nansum(synth_site_entry.loc[synth_site_entry["YEAR"] == year, "PRECIP"].values))
        logbins = np.logspace(np.nanmin(np.log10(obs_annuals)), np.nanmax(np.log10(obs_annuals)), 10)
        axis.set(title=sites[a])
        axis.set(xscale="log")
        axis.hist(synth_annuals, density=True, bins=logbins, color="grey", alpha=1)
        axis.hist(obs_annuals, density=True, bins=logbins, color="black", histtype="step")
        if a == 0: axis.legend(["Obs", "Synth"])
    plt.tight_layout()
    comp_gmmhmm_fig.savefig("{}Compare_GMMHMM_AnnualPrecip.{}".format(dp, ext))
    plt.close()
     
    # cumulative frequency of precipitation plot
    comp_cumfreq_fig, axes = plt.subplots(nrows=r_sites, ncols=c_sites, figsize=(16, 9), sharex="all", sharey="all")
    comp_cumfreq_fig.suptitle("Cumulative Frequency Curves for Precipitation")
    comp_cumfreq_fig.supxlabel("% Exceedance")
    comp_cumfreq_fig.supylabel("Daily Precipitation [m]" if resolution == "daily" else "Monthly Precipitation [m]")
    for a, axis in enumerate(axes.flat):
        axis.set(title=sites[a])
        axis.set(yscale="log", ylim=[1E-5, 1E-1]) if resolution == "daily" else axis.set(yscale="log", ylim=[2.54E-4, 1])
        obs_site_idx, synth_site_idx = obs_df["SITE"] == sites[a], synth_df["SITE"] == sites[a]
        obs_site_entry, synth_site_entry = obs_df.loc[obs_site_idx], synth_df.loc[synth_site_idx]
        descending_obs = np.array(sorted(obs_site_entry["PRECIP"].values, reverse=True))
        descending_synth = np.array(sorted(synth_site_entry["PRECIP"].values, reverse=True))
        obs_exceedance = np.array([np.nansum(descending_obs >= point) / len(descending_obs) for point in descending_obs])
        synth_exceedance = np.array([np.nansum(descending_synth >= point) / len(descending_synth) for point in descending_synth])
        axis.plot(100 * obs_exceedance, descending_obs, c="black", label="Obs")
        axis.plot(100 * synth_exceedance, descending_synth, c="grey", linestyle="-.", label="Synth")
        if a == 0: axis.legend()
    plt.tight_layout()
    comp_cumfreq_fig.savefig("{}Compare_CumulativeFrequency_Precip.{}".format(dp, ext))
    plt.close()
    
    # spatial correlations for precip and temp
    obs_monthly_dict, synth_monthly_dict = {}, {}
    for month in sorted(set(obs_df["MONTH"].values)):
        obs_monthly_dict[int(month)] = {"PRECIP": pd.DataFrame(), "TEMP": pd.DataFrame()}
        obs_month_idx = obs_df["MONTH"] == month
        obs_month_entry = obs_df.loc[obs_month_idx]
        for site in sorted(set(obs_month_entry["SITE"].values)):
            obs_site_idx = obs_month_entry["SITE"] == site
            obs_site_entry = obs_month_entry.loc[obs_site_idx]
            obs_monthly_dict[month]["PRECIP"][site] = obs_site_entry["PRECIP"].values
            obs_monthly_dict[month]["TEMP"][site] = obs_site_entry["TEMP"].values    
    for month in sorted(set(synth_df["MONTH"].values)):
        synth_monthly_dict[int(month)] = {"PRECIP": pd.DataFrame(), "TEMP": pd.DataFrame()}
        synth_month_idx = synth_df["MONTH"] == month
        synth_month_entry = synth_df.loc[synth_month_idx]
        for site in sorted(set(synth_month_entry["SITE"].values)):
            synth_site_idx = synth_month_entry["SITE"] == site
            synth_site_entry = synth_month_entry.loc[synth_site_idx]
            synth_monthly_dict[month]["PRECIP"][site] = synth_site_entry["PRECIP"].values
            synth_monthly_dict[month]["TEMP"][site] = synth_site_entry["TEMP"].values
    corr_cmap = plt.get_cmap("gnuplot", 21)
    for month in obs_monthly_dict.keys():
        compare_spatial_fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(16, 9), sharex="all", sharey="all")
        compare_spatial_fig.suptitle("Compare Spatial Correlations: {}".format(month_names[month-1]))
        for i, wvar in enumerate(["PRECIP", "TEMP"]):
            obs_axis, synth_axis = axes[i, 0], axes[i, 1]
            obs_axis.set_title("Obs {}".format(wvar.title())), synth_axis.set_title("Synth {}".format(wvar.title()))
            obs_corr = obs_monthly_dict[month][wvar].corr(method="pearson") 
            synth_corr = synth_monthly_dict[month][wvar].corr(method="pearson")
            obs_axis.imshow(obs_corr.values, vmin=0, vmax=1, cmap=corr_cmap)
            for j in range(obs_corr.shape[0]):
                for k in range(obs_corr.shape[1]):
                    obs_axis.text(k, j, '{:.2f}'.format(obs_corr.values[j, k]), ha='center', va='center', backgroundcolor=[0, 0, 0, 0.25], color="white")
            corr_colors = synth_axis.imshow(synth_corr.values, vmin=0, vmax=1, cmap=corr_cmap)
            for j in range(synth_corr.shape[0]):
                for k in range(synth_corr.shape[1]):
                    synth_axis.text(k, j, '{:.2f}'.format(synth_corr.values[j, k]), ha='center', va='center', backgroundcolor=[0, 0, 0, 0.25], color="white")
            obs_axis.set_xticks(range(len(sites))), obs_axis.set_yticks(range(len(sites)))
            obs_axis.set_xticklabels(sites, rotation=45, ha="right"), obs_axis.set_yticklabels(sites)
            synth_axis.set_xticks(range(len(sites))), synth_axis.set_yticks(range(len(sites)))
            synth_axis.set_xticklabels(sites, rotation=45, ha="right"), synth_axis.set_yticklabels(sites)
        compare_spatial_fig.colorbar(corr_colors, ax=axes[:, 1], label="Pearson Correlation Coeff [-]")
        compare_spatial_fig.savefig("{}Compare_SpatialCorrelations_{}.{}".format(dp, month_names[month-1], ext))
        plt.close()
        
    # monthly temporal correlations for precip and temp
    for s, site in enumerate(sites):
        compare_temporal_fig, axes = plt.subplots(nrows=4, ncols=2, figsize=(9, 16), sharex="all", sharey="row")
        if resolution == "monthly":
            compare_temporal_fig.suptitle("Compare Monthly Temporal Correlations: {} | Obs and Synth should have same shape and magnitude".format(site))
        else:    
            compare_temporal_fig.suptitle("Compare Daily Temporal Correlations: {} | Obs and Synth should have same shape and magnitude".format(site))
        compare_temporal_fig.supxlabel("Lag")
        obs_precip_acf_axis, synth_precip_acf_axis = axes[0, 0], axes[0, 1]
        obs_precip_pacf_axis, synth_precip_pacf_axis = axes[1, 0], axes[1, 1]
        obs_temp_acf_axis, synth_temp_acf_axis = axes[2, 0], axes[2, 1]
        obs_temp_pacf_axis, synth_temp_pacf_axis = axes[3, 0], axes[3, 1]
        obs_site_idx, synth_site_idx = obs_df["SITE"] == site, synth_df["SITE"] == site
        obs_site_entry, synth_site_entry = obs_df.loc[obs_site_idx], synth_df.loc[synth_site_idx]
        plot_acf(ax=obs_precip_acf_axis, x=obs_site_entry["PRECIP"].values, use_vlines=False, color="royalblue")
        plot_acf(ax=synth_precip_acf_axis, x=synth_site_entry["PRECIP"].values, use_vlines=False, color="royalblue")
        plot_pacf(ax=obs_precip_pacf_axis, x=obs_site_entry["PRECIP"].values, use_vlines=False, color="royalblue")
        plot_pacf(ax=synth_precip_pacf_axis, x=synth_site_entry["PRECIP"].values, use_vlines=False, color="royalblue")
        plot_acf(ax=obs_temp_acf_axis, x=obs_site_entry["TEMP"].values, use_vlines=False, color="firebrick")
        plot_acf(ax=synth_temp_acf_axis, x=synth_site_entry["TEMP"].values, use_vlines=False, color="firebrick")
        plot_pacf(ax=obs_temp_pacf_axis, x=obs_site_entry["TEMP"].values, use_vlines=False, color="firebrick")
        plot_pacf(ax=synth_temp_pacf_axis, x=synth_site_entry["TEMP"].values, use_vlines=False, color="firebrick")
        obs_precip_acf_axis.set_title("Obs"), synth_precip_acf_axis.set_title("Synth")
        for axis in [obs_precip_pacf_axis, synth_precip_pacf_axis, obs_temp_acf_axis, synth_temp_acf_axis, obs_temp_pacf_axis, synth_temp_pacf_axis]:
            axis.set_title("")
        obs_precip_acf_axis.set_ylabel("Precip ACF [-]"), obs_temp_acf_axis.set_ylabel("Temp ACF [-]")
        obs_precip_pacf_axis.set_ylabel("Precip PACF [-]"), obs_temp_pacf_axis.set_ylabel("Temp PACF [-]")
        obs_temp_pacf_axis.set_xlim([0, 24])
        plt.tight_layout()
        compare_temporal_fig.savefig("{}Compare_TemporalCorrelations_{}.{}".format(dp, site.replace(" ", ""), ext))
        plt.close()

    # compare the observed and synthetic Kendall/Spearman correlation coefficients
    compare_ks_fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(16, 9), sharex="all", sharey="all")
    compare_ks_fig.suptitle("Correlation of Spatially Averaged P/T Data by Month")
    compare_ks_fig.supxlabel("Month"), compare_ks_fig.supylabel("Correlation Coefficient [-]") 
    obs_ks_corrs_dict, synth_ks_corrs_dict = {"Kendall": [], "Spearman": []}, {"Kendall": [], "Spearman": []}
    for month in obs_monthly_dict.keys():
        obs_month_idx, synth_month_idx = obs_df["MONTH"] == month, synth_df["MONTH"] == month
        obs_month_entry, synth_month_entry = obs_df.loc[obs_month_idx], synth_df.loc[synth_month_idx]
        obs_ks_dict = {"PRECIP": [], "TEMP": []}
        synth_ks_dict = {"PRECIP": [], "TEMP": []}
        for year in sorted(set(obs_month_entry["YEAR"].values)):
            obs_year_idx = obs_month_entry["YEAR"] == year
            obs_year_entry = obs_month_entry.loc[obs_year_idx]
            if resolution == "monthly":
                obs_ks_dict["PRECIP"].append(np.nanmean(obs_year_entry["PRECIP"].values))
                obs_ks_dict["TEMP"].append(np.nanmean(obs_year_entry["TEMP"].values))
            else:
                agg_obs_precip, agg_obs_temp = [], []
                for site in sorted(set(obs_year_entry["SITE"].values)):
                    agg_obs_precip.append(np.nansum(obs_year_entry.loc[obs_year_entry["SITE"] == site, "PRECIP"].values))
                    agg_obs_temp.append(np.nanmean(obs_year_entry.loc[obs_year_entry["SITE"] == site, "TEMP"].values))
                obs_ks_dict["PRECIP"].append(np.nanmean(agg_obs_precip))
                obs_ks_dict["TEMP"].append(np.nanmean(agg_obs_temp))
        for year in sorted(set(synth_month_entry["YEAR"].values)):
            synth_year_idx = synth_month_entry["YEAR"] == year
            synth_year_entry = synth_month_entry.loc[synth_year_idx]
            if resolution == "monthly":
                synth_ks_dict["PRECIP"].append(np.nanmean(synth_year_entry["PRECIP"].values))
                synth_ks_dict["TEMP"].append(np.nanmean(synth_year_entry["TEMP"].values))
            else:
                agg_synth_precip, agg_synth_temp = [], []
                for site in sorted(set(synth_year_entry["SITE"].values)):
                    agg_synth_precip.append(np.nansum(synth_year_entry.loc[synth_year_entry["SITE"] == site, "PRECIP"].values))
                    agg_synth_temp.append(np.nanmean(synth_year_entry.loc[synth_year_entry["SITE"] == site, "TEMP"].values))
                synth_ks_dict["PRECIP"].append(np.nanmean(agg_synth_precip))
                synth_ks_dict["TEMP"].append(np.nanmean(agg_synth_temp))
        obs_ks_df = pd.DataFrame({"P": obs_ks_dict["PRECIP"], "T": obs_ks_dict["TEMP"]}) 
        synth_ks_df = pd.DataFrame({"P": synth_ks_dict["PRECIP"], "T": synth_ks_dict["TEMP"]}) 
        obs_ks_corrs_dict["Kendall"].append(obs_ks_df.corr(method="kendall")["P"]["T"])
        obs_ks_corrs_dict["Spearman"].append(obs_ks_df.corr(method="spearman")["P"]["T"])
        synth_ks_corrs_dict["Kendall"].append(synth_ks_df.corr(method="kendall")["P"]["T"])
        synth_ks_corrs_dict["Spearman"].append(synth_ks_df.corr(method="spearman")["P"]["T"])
    for a, axis in enumerate(axes.flat):
        if a == 0: 
            ks_corrs_dict = obs_ks_corrs_dict
            axis.set_title("Obs")
            ks_color = "black"
        if a == 1:
            ks_corrs_dict = synth_ks_corrs_dict
            axis.set_title("Synth")
            ks_color = "grey"
        axis.plot(range(len(obs_monthly_dict.keys())), ks_corrs_dict["Kendall"], color=ks_color, marker="s", label=r"Kendall $\tau$", zorder=10)
        axis.plot(range(len(obs_monthly_dict.keys())), ks_corrs_dict["Spearman"], color=ks_color, marker="d", label=r"Spearman $\rho$", zorder=10)
        axis.legend()
        axis.hlines(0, 0, len(month_names)-1, color="red", linestyle="dashed", zorder=9) 
        axis.set_ylim(-1, 1)
        axis.set_xticks(range(len(month_names)))
        axis.set_xticklabels(month_names, rotation=45)
    plt.tight_layout()
    compare_ks_fig.savefig("{}Compare_PTCorrelations_KendallSpearman.{}".format(dp, ext))
    plt.close()

    # compare via scatterplot and histogram the monthly/daily observed and synthetic
    for site in sites:
        obs_site_idx, synth_site_idx = obs_df["SITE"] == site, synth_df["SITE"] == site
        compare_histscatter_fig = plt.figure(figsize=(16, 9))
        if resolution == "monthly":
            compare_histscatter_fig.suptitle("Direct Comparison of Montly Precipitation and Temperature at {} | PRECIP on x-axis, TEMP on y-axis".format(site))
        else:
            compare_histscatter_fig.suptitle("Direct Comparison of Daily Precipitation and Temperature at {} | PRECIP on x-axis, TEMP on y-axis".format(site))
        sub_figs = compare_histscatter_fig.subfigures(3, 4)
        for i, sub_fig in enumerate(sub_figs.flat):
            obs_month_idx, synth_month_idx = obs_df["MONTH"] == i+1, synth_df["MONTH"] == i+1
            axes = sub_fig.subplots(2, 2, gridspec_kw={"width_ratios": [4, 1], "height_ratios": [1, 3]})
            sub_fig.subplots_adjust(wspace=0, hspace=0)
            synthColor = "grey"
            for j, axis in enumerate(axes.flat):
                if j == 0:
                    # precip histogram comparison
                    axis.hist(synth_df.loc[synth_site_idx & synth_month_idx, "PRECIP"].values, density=True, color="grey")
                    axis.hist(obs_df.loc[obs_site_idx & obs_month_idx, "PRECIP"].values, density=True, color="black", histtype="step")
                    axis.set(xticks=[], yticks=[])
                if j == 1:
                    # label for the month
                    axis.axis("off")
                    axis.text(0.5, 0.5, month_names[i], transform=axis.transAxes, va="center", ha="center")
                if j == 2:
                    # scatterplot
                    axis.scatter(synth_df.loc[synth_site_idx & synth_month_idx, "PRECIP"].values,
                                 synth_df.loc[synth_site_idx & synth_month_idx, "TEMP"].values,
                                 marker=".", facecolors="grey", alpha=0.2, rasterized=True)
                    axis.scatter(obs_df.loc[obs_site_idx & obs_month_idx, "PRECIP"].values, 
                                 obs_df.loc[obs_site_idx & obs_month_idx, "TEMP"].values, 
                                 marker="o", facecolors="none", edgecolors="black")
                if j == 3:
                    # temp histogram comparison
                    axis.hist(synth_df.loc[synth_site_idx & synth_month_idx, "TEMP"].values, density=True, color="grey", orientation="horizontal")
                    axis.hist(obs_df.loc[obs_site_idx & obs_month_idx, "TEMP"].values, density=True, color="black", histtype="step", orientation="horizontal")
                    axis.set(xticks=[], yticks=[])
        compare_histscatter_fig.savefig("{}Compare_HistScatter_{}.{}".format(dp, site.replace(" ", ""), ext))
        plt.close()

    # statistical tests for the observed/synthetic precipitation/temperature distributions
    # -- Mann-Whitney U (MWU), --> median
    # -- Levene --> std
    # -- Kolmogorov-Smirnov (T_n) --> empirical distribution sensitive to mean
    goodness_stats, bar_width = ["MWU", "Levene", "$T_n$"], 0.5
    r_months, c_months = squarest_subplots(len(month_names))
    for site in sites:
        obs_site_idx, synth_site_idx = obs_df["SITE"] == site, synth_df["SITE"] == site
        stats_fig, axes = plt.subplots(nrows=r_months, ncols=c_months, figsize=(16, 9), sharex="all", sharey="all")
        if resolution == "monthly":
            stats_fig.suptitle("Null-Hypothesis Monthly Statistics at {} | Mann-Whitney U, Levene, Kolmogorov-Smirnov".format(site))
        else:
            stats_fig.suptitle("Null-Hypothesis Daily Statistics at {} | Mann-Whitney U, Levene, Kolmogorov-Smirnov".format(site))
        stats_fig.supylabel("p-Value [-]")
        for i, axis in enumerate(axes.flat):
            obs_month_idx, synth_month_idx = obs_df["MONTH"] == i+1, synth_df["MONTH"] == i+1
            obs_precips = obs_df.loc[obs_site_idx & obs_month_idx, "PRECIP"].astype(float).values 
            obs_temps = obs_df.loc[obs_site_idx & obs_month_idx, "TEMP"].astype(float).values
            synth_precips = synth_df.loc[synth_site_idx & synth_month_idx, "PRECIP"].astype(float).values 
            synth_temps = synth_df.loc[synth_site_idx & synth_month_idx, "TEMP"].astype(float).values
            prcpMWU, tavgMWU = scipy.stats.mannwhitneyu(obs_precips, synth_precips, nan_policy="omit"), scipy.stats.mannwhitneyu(obs_temps, synth_temps, nan_policy="omit")
            prcpLevene, tavgLevene = scipy.stats.levene(obs_precips, synth_precips[np.isfinite(synth_precips)]), scipy.stats.levene(obs_temps, synth_temps[np.isfinite(synth_temps)])
            prcpKS, tavgKS = scipy.stats.ks_2samp(obs_precips, synth_precips[np.isfinite(synth_precips)]), scipy.stats.ks_2samp(obs_temps, synth_temps[np.isfinite(synth_temps)])
            axis.set(title=month_names[i], xlim=[-bar_width * 1.5, 2 * (len(goodness_stats) - 1 + bar_width)], ylim=[0, 1])
            axis.bar([2 * x - bar_width / 2 for x in range(len(goodness_stats))], [prcpMWU[1], prcpLevene[1], prcpKS.pvalue], 
                     width=bar_width, color="royalblue", zorder=10, label="Precip")
            axis.bar([2 * x + bar_width / 2 for x in range(len(goodness_stats))], [tavgMWU[1], tavgLevene[1], tavgKS.pvalue], 
                     width=bar_width, color="orangered", zorder=10, label="Temp")
            if i == 11:
                axis.legend().set_zorder(11)
            axis.hlines(0.05, -2 * bar_width, 2 * (len(goodness_stats) + bar_width), color="black", linestyles="dashed", zorder=11)
            axis.set_xticks([2 * x for x in range(len(goodness_stats))])
            axis.set_xticklabels(labels=goodness_stats)
        plt.tight_layout()
        stats_fig.savefig("{}Compare_StatisticalDistributions_{}.{}".format(dp, site.replace(" ", ""), ext))
        plt.close()

    # if daily, plot distribution by DOY
    if resolution == "daily":
        for site in sites:
            obs_site_idx, synth_site_idx = obs_df["SITE"] == site, synth_df["SITE"] == site
            obs_site_entry, synth_site_entry = obs_df.loc[obs_site_idx], synth_df.loc[synth_site_idx]
            doy_fig, axes = plt.subplots(nrows=2, ncols=1, figsize=(16, 9), sharex="all")
            doy_fig.suptitle("Comparison of Obs to Synth by DOY, with Medians (black, grey lines) and 5% - 95% Percentiles (color, grey bands)")
            doy_fig.supxlabel("DOY")
            obs_doy_dict, synth_doy_dict = {}, {}
            for i in range(obs_site_entry.shape[0]):
                obs_i_entry = obs_site_entry.iloc[i]
                doy = int(dt.datetime.strptime("{}-{}-{}".format(str(obs_i_entry["YEAR"]).zfill(4), str(obs_i_entry["MONTH"]).zfill(2), str(obs_i_entry["DAY"]).zfill(2)), 
                                               "%Y-%m-%d").strftime("%j"))
                if doy not in obs_doy_dict:
                    obs_doy_dict[doy] = {"PRECIP": [obs_i_entry["PRECIP"]], "TEMP": [obs_i_entry["TEMP"]]}
                else:
                    obs_doy_dict[doy]["PRECIP"].append(obs_i_entry["PRECIP"])
                    obs_doy_dict[doy]["TEMP"].append(obs_i_entry["TEMP"])
            for i in range(synth_site_entry.shape[0]):
                synth_i_entry = synth_site_entry.iloc[i]
                if synth_i_entry["MONTH"] == 2 and synth_i_entry["DAY"] == 29: continue
                doy = int(dt.datetime.strptime("{}-{}-{}".format(str(synth_i_entry["YEAR"]).zfill(4), str(synth_i_entry["MONTH"]).zfill(2), str(synth_i_entry["DAY"]).zfill(2)), 
                                               "%Y-%m-%d").strftime("%j"))
                if doy not in synth_doy_dict:
                    synth_doy_dict[doy] = {"PRECIP": [float(synth_i_entry["PRECIP"])], "TEMP": [float(synth_i_entry["TEMP"])]}
                else:
                    synth_doy_dict[doy]["PRECIP"].append(float(synth_i_entry["PRECIP"]))
                    synth_doy_dict[doy]["TEMP"].append(float(synth_i_entry["TEMP"]))
            for a, axis in enumerate(axes.flat):
                if a == 0:
                    wvar = "PRECIP"
                    wvcolor = "royalblue"
                    axis.set_ylabel(r"Non-Zero Precipitation [m]")
                    axis.set_yscale("log")
                    axis.set_ylim(bottom=0.00001)
                if a == 1:
                    wvar = "TEMP"
                    wvcolor = "firebrick"
                    axis.set_ylabel("Temperature ["+chr(176)+"C]")
                wvar = "PRECIP" if a == 0 else "TEMP"
                obs_p5, obs_p50, obs_p95 = [], [], []
                synth_p5, synth_p50, synth_p95 = [], [], []
                for d in obs_doy_dict.keys():
                    obs = np.array(obs_doy_dict[d][wvar])
                    obs = obs[obs > 0] if wvar == "PRECIP" else obs
                    obs_p5.append(np.nanpercentile(obs, 5))
                    obs_p50.append(np.nanpercentile(obs, 50))
                    obs_p95.append(np.nanpercentile(obs, 95))
                for d in synth_doy_dict.keys():
                    synth = np.array(synth_doy_dict[d][wvar])
                    synth = synth[synth > 0] if wvar == "PRECIP" else synth
                    synth_p5.append(np.nanpercentile(synth, 5))
                    synth_p50.append(np.nanpercentile(synth, 50))
                    synth_p95.append(np.nanpercentile(synth, 95))
                synth_doys, obs_doys = np.array(list(synth_doy_dict.keys())), np.array(list(obs_doy_dict.keys()))
                sorted_synth_doys = synth_doys[np.argsort(synth_doys)]
                sorted_synth_p5 = np.array(synth_p5)[np.argsort(synth_doys)]
                sorted_synth_p50 = np.array(synth_p50)[np.argsort(synth_doys)]
                sorted_synth_p95 = np.array(synth_p95)[np.argsort(synth_doys)]
                sorted_obs_doys = obs_doys[np.argsort(obs_doys)]
                sorted_obs_p5 = np.array(obs_p5)[np.argsort(obs_doys)]
                sorted_obs_p50 = np.array(obs_p50)[np.argsort(obs_doys)]
                sorted_obs_p95 = np.array(obs_p95)[np.argsort(obs_doys)]
                axis.fill_between(sorted_synth_doys, sorted_synth_p5, sorted_synth_p95, color="grey", alpha=0.33, zorder=10)
                axis.fill_between(sorted_obs_doys, sorted_obs_p5, sorted_obs_p95, color=wvcolor, alpha=0.33, zorder=11)
                axis.plot(sorted_synth_doys, sorted_synth_p50, color="grey", linestyle="-.", zorder=12)
                axis.plot(sorted_obs_doys, sorted_obs_p50, color="black", zorder=12)
            plt.tight_layout()
            doy_fig.savefig("{}Compare_PerDOY_{}.{}".format(dp, site.replace(" ", ""), ext))
            plt.close()



from typing import List, Tuple, Dict 
import numpy as np
import pandas as pd
import datetime as dt
import random
from hmmlearn.hmm import GMMHMM
import warnings
from statsmodels.tsa.ar_model import AutoReg
from copulas import univariate, bivariate, multivariate
import scipy
import copulae
from statsmodels.tools import eval_measures
import logging

from .make_figures import *


def fit_data(data: pd.DataFrame, 
             resolution: str,
             validation: bool,
             dirpath: str,
             fit_kwargs: dict) -> list[dict, dict]:
    """
    Managing function that fits the raw climate/weather data as a reformatted DataFrame,
    with statistics and sampling schemes for precipitation, additional parameters
    (i.e. temperature), and the relationship between the two

    Parameters
    ----------
    data: pd.DataFrame
        Temporally reformatted data to be used for the fitting
    resolution: str
        The temporal resolution of the input data. Can be 'monthly' or 'daily'
    validation: bool
        Flag for producing figures to validate each step of the generator
    dirpath: str
        Path for where to save the validation figures
    fit_kwargs: dict
        Dictionary with the fit keyword arguments

    Returns
    -------
    precip_fit_dict: dict
        Dictionary containing statistical information related to fitting of precipitation data
    copulaetemp_fit_dict: dict
        Dictionary containing statistical information related to conditional fitting of temperature 
        data through copulae
    """
   
    # fit kwargs
    default_fit_kwargs = {"gmmhmm_min_states": 1,
                          "gmmhmm_max_states": 4,
                          "gmmhmm_states": 0,
                          "ar_lag": 1,
                          "stationarity_groups": 2,
                          "copula_families": ["Independence", "Frank", "Gaussian"],
                          "figure_extension": "svg",
                          "validation_figures": ["precip", "copula"],
                          "fit_verbose": True}
    if not fit_kwargs: 
        fit_kwargs = default_fit_kwargs
    else:
        for k in default_fit_kwargs:
            if k not in fit_kwargs:
                fit_kwargs[k] = default_fit_kwargs[k] 
    
    # validation
    global do_validation, validation_dirpath, validation_extension, validation_figures, fit_verbose
    do_validation, validation_dirpath, validation_extension, validation_figures = validation, dirpath, fit_kwargs["figure_extension"], fit_kwargs["validation_figures"]
    fit_verbose = fit_kwargs["fit_verbose"]
 
    # precip
    precip_col_idx = list(data.columns).index("PRECIP")
    precip_fit_dict = fit_precip(data[data.columns[:precip_col_idx+1]].copy(), 
                                 resolution,
                                 fit_kwargs["gmmhmm_min_states"],
                                 fit_kwargs["gmmhmm_max_states"],
                                 fit_kwargs["gmmhmm_states"])

    # copulae/temp
    copulaetemp_fit_dict = fit_copulae(data[data.columns[:precip_col_idx+1].append(pd.Index(["TEMP"]))].copy(), 
                                       resolution, 
                                       precip_fit_dict["log10_annual_precip"].index.values,
                                       fit_kwargs["ar_lag"], 
                                       fit_kwargs["stationarity_groups"],
                                       fit_kwargs["copula_families"])
    
    # validation for fits
    if do_validation:
        validate_pt_fits(validation_dirpath, validation_extension, data, precip_fit_dict, copulaetemp_fit_dict, validation_figures)
    
    return precip_fit_dict, copulaetemp_fit_dict


def fit_precip(data: pd.DataFrame, resolution: str, min_states: int, max_states: int, fixed_states: int) -> dict:
    """
    Function that fits and validates the precipitation data. Precipitation is transformed
    to a log-scale, annualized (summed), and fit to a Gaussian mixture-model Hidden 
    Markov model (GMMHMM)
    
    Parameters
    ----------
    data: pd.DataFrame
        Temporal reformat of ``raw_data`` where each year, month, day have their own column
        in the DataFrame
    resolution: str
        The temporal resolution of the input data. Can be 'monthly' or 'daily' 
    min_states: int
        The minimum number of hidden states to try fitting 
    max_states: int
        The maximum number of hidden states to try fitting. More than ~6 tends
        to perform poorly in terms of best fit and length of computation 
    fixed_states: int
        Do not try fitting for best number of hidden states and only use this value
    
    Returns
    -------
    precip_fit_dict: dict
        Dictionary containing statistical information related to fitting of precipitation data
    """

    def gmmhmm_state_num_estimator(transformed_df: pd.DataFrame, lengths: list[int], min_states: int, max_states: int, iterations: int = 10) -> int:
        """
        Function to programmatically determine the best-fitting number of states 
        for the Gaussian mixture model hidden Markov model
        
        Parameters
        ----------
        transformed_df: pd.DataFrame
            DataFrame for the log10-transformed annualized precipitation values, with
            associated sites and years organized as a DataFrame where indices are
            years and columns are sites
        lengths: list[int]
            Length of each sequence of consecutive years in the data
        min_states: int
            The minimum number of hidden states to try fitting. Default: 1 
        max_states: int
            The maximum number of hidden states to try fitting. More than ~6 tends
            to perform poorly in terms of best fit and length of computation. Default: 5 
        iterations: int, fixed
            The number of attempts to find a best-fitting model for each unique
            number of states -- this is necessary because the convergence EM method can 
            fall into local optima. From all iterations, the one with the highest 
            log-likelihood is used. Using: 10

        Returns
        -------
        best_num_states: int
            Best-fitting model's number of states, as determined by Bayesian Information
            Criterion (BIC). Log-likelihood and Akaike Information Criterion (AIC) are 
            also used, but log-likelihood is monotonically increasing with number of states
            and (from experimentation) AIC does not penalize additional states strongly enough. 
        """

        n_stations = len(transformed_df.columns)
        models, seeds, AICs, BICs, LLs = [], [], [], [], []
        max_attempts = 50
        for num_states in range(min_states, max_states + 1):
            best_model, best_LL, best_seed = None, None, None
            for _ in range(iterations):
                positive_definite, temp_seed, temp_model = False, None, None
                attempts = 0
                while (not positive_definite) and (attempts < max_attempts):
                    # get the random state
                    tempSeed = random.getstate()

                    # define the parameters for the model
                    temp_model_inst = GMMHMM(n_components=num_states, n_iter=1000, covariance_type="full", init_params="cmw", verbose=False)
                    temp_model_inst.startprob_ = np.full(shape=num_states, fill_value=1./num_states)
                    temp_model_inst.transmat_ = np.full(shape=(num_states, num_states), fill_value=1./num_states)
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore", category=RuntimeWarning)
                        temp_model = temp_model_inst.fit(transformed_df.values, lengths=lengths)

                    # get, reshape the covariance matrices
                    temp_covars = temp_model.covars_.reshape((num_states, n_stations, n_stations))

                    # check that each state has a covariance matrix that is positive definite, meaning:
                    # -- it's symmetric
                    symmetric_check = all([np.allclose(temp_covars[i].T, temp_covars[i]) for i in range(num_states)])
                    # -- it's eigenvalues are all positive
                    eig_check = all([(np.linalg.eigvalsh(temp_covars[i]) > 0).all() for i in range(num_states)])

                    # state if the covariance matrix is positive definite or not
                    positive_definite = symmetric_check and eig_check
                    attempts += 1

                # calculate the loglikelihood of this model, save best
                if attempts < max_attempts: 
                    temp_score = temp_model.score(transformed_df.values, lengths=lengths)
                    if not best_LL or temp_score > best_LL:
                        best_LL = temp_score
                        best_model = temp_model
                        best_seed = temp_seed
            # save the best metrics for each number of states
            if attempts < max_attempts:
                models.append(best_model)
                seeds.append(best_seed)
                LLs.append(best_LL)
                AICs.append(best_model.aic(transformed_df.values, lengths=lengths))
                BICs.append(best_model.bic(transformed_df.values, lengths=lengths))
                if fit_verbose:
                    print("Positive definite covariance matrix for GMMHMM fit found for {} state(s)!".format(num_states))
            else:
                if fit_verbose:
                    print("Positive definite covariance matrix for GMMHMM fit cannot be found for {} states...".format(num_states))
        model = models[np.argmin(BICs)]
        
        # validate if prompted
        if do_validation and ("precip" in validation_figures):
            validate_gmmhmm_states(validation_dirpath, validation_extension, min_states, max_states, LLs, AICs, BICs)
        
        return model.n_components
    

    def conceptual_reorder(model_means: np.array, model_covars: np.array, model_transmat: np.array) -> tuple[np.array, np.array, np.array]:
        """
        Function to conceptually reorder the discovered ``n`` states, with the ``0``th index
        being the driest state and the (``n-1``)th index being the wettest state

        Parameters
        ----------
        model_means: np.array
            Means discovered by the GMMHMM, ordered as discovered
        model_covars: np.array
            Covariance matrix discovered by the GMMHMM, ordered as discovered
        model_transmat: np.array
            Transition matrix discovered by the GMMHMM, ordered as discovered
        
        Returns
        -------
        reordered_means: np.array
            Means reordered so that the driest state is the first index and the 
            wettest is the last
        reordered_covars: np.array
            Means reordered so that the driest state is the first index and the 
            wettest is the last
        reordered_transmat: np.array
            Means reordered so that the driest state is the first index and the 
            wettest is the last
        """
        
        # reorder indices, return if already in order
        conceptual_order_indices = np.argsort([np.sum(model_means[i]) for i in range(len(model_means))])
        if np.all(np.diff(conceptual_order_indices) >= 0):
            return model_means, model_covars, model_transmat

        # reorder means, covariances
        reordered_means = model_means[conceptual_order_indices, :]
        reordered_covars = model_covars[conceptual_order_indices, :]

        # reorder transmat
        reordered_transmat = np.full(shape=model_transmat.shape, fill_value=np.nan)
        for i in range(model_transmat.shape[0]):
            reordered_transmat[conceptual_order_indices[i], :] = model_transmat[i, conceptual_order_indices]

        # return the reordered matrices
        return reordered_means, reordered_covars, reordered_transmat
    
    
    logging.getLogger("hmmlearn.base").setLevel(logging.ERROR)
    # annualize precipitation, log10 transformation, format so index=years, column=sites, cell=precip
    sites, years = sorted(set(data["SITE"].values)), sorted(set(data["YEAR"].values))
    transformed_data = pd.DataFrame(index=years, columns=sites)
    for site in sites:
        site_idx = data["SITE"] == site
        site_entry = data[site_idx]
        site_years = sorted(set(site_entry["YEAR"].values))
        for year in years:
            year_idx = site_entry["YEAR"] == year
            year_entry = site_entry[year_idx]
            if year_entry.empty: continue
            transformed_data.at[year, site] = np.log10(np.sum(year_entry["PRECIP"].values))
    transformed_data.astype({col: float for col in sites})
    transformed_data.dropna(inplace=True)

    # checking if there are any years missing from the data, making sequences of existing years
    seq_lengths, l = [], 0
    for year in years:
        if year in list(transformed_data.index):
            l += 1
        else:
            seq_lengths.append(l)
            l = 0
    seq_lengths.append(l)
    seq_lengths = [l for l in seq_lengths if l != 0]

    # determine best-fitting number of states for the GMMHMM
    if fixed_states > 0:
        num_states = fixed_states
    else:
        num_states = gmmhmm_state_num_estimator(transformed_data, seq_lengths, min_states=min_states, max_states=max_states)
    
    # use these num_states to fit the GMMHMM
    best_model, best_LL, best_seed = None, None, None
    for _ in range(20):
        positive_definite, temp_seed, temp_model = False, None, None
        while not positive_definite:
            tempSeed = random.getstate()
            temp_model_inst = GMMHMM(n_components=num_states, n_iter=1000, covariance_type="full", init_params="cmw", verbose=False)
            temp_model_inst.startprob_ = np.full(shape=num_states, fill_value=1./num_states)
            temp_model_inst.transmat_ = np.full(shape=(num_states, num_states), fill_value=1./num_states)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                temp_model = temp_model_inst.fit(transformed_data.values, lengths=seq_lengths)
            temp_covars = temp_model.covars_.reshape((num_states, len(sites), len(sites)))
            symmetric_check = all([np.allclose(temp_covars[i].T, temp_covars[i]) for i in range(num_states)])
            eig_check = all([(np.linalg.eigvalsh(temp_covars[i]) > 0).all() for i in range(num_states)])
            positive_definite = symmetric_check and eig_check

        temp_score = temp_model.score(transformed_data.values, lengths=seq_lengths)
        if not best_LL or temp_score > best_LL:
            best_LL = temp_score
            best_model = temp_model
            best_seed = temp_seed
    seed, model = best_seed, best_model
    model.means_, model.covars_, model.transmat_ = conceptual_reorder(model.means_, model.covars_, model.transmat_)
 
    # predict the hidden state for each datapoint
    hidden_states = model.predict(transformed_data.values, lengths=seq_lengths)

    # extract means, covariance, standard deviation from model
    means = model.means_.reshape((model.n_components, len(sites)))
    masked_covars = model.covars_.copy()
    masked_covars[masked_covars < 0] = np.nan
    stds = np.sqrt(masked_covars).reshape((model.n_components, len(sites), len(sites)))
    stds = np.array([[stds[n][i][i] for i in range(len(sites))] for n in range(num_states)])

    # return the model and associated statistics as a dictionary
    precip_fit_dict = {}
    precip_fit_dict["log10_annual_precip"] = transformed_data
    precip_fit_dict["seq_lengths"] = seq_lengths
    precip_fit_dict["num_gmmhmm_states"] = num_states
    precip_fit_dict["seed"] = seed
    precip_fit_dict["model"] = model
    precip_fit_dict["means"] = means
    precip_fit_dict["covars"] = masked_covars
    precip_fit_dict["stds"] = stds
    precip_fit_dict["hidden_states"] = hidden_states
    precip_fit_dict["t_probs"] = model.transmat_
    return precip_fit_dict


def fit_copulae(data: pd.DataFrame, resolution: str, precip_fit_years: list[int], 
                ar_lag: int, stationarity_groups: int, copula_families: list[str]) -> dict:
    """
    Function that fits and validates the temperature data through fitting of 
    hydroclimatic copulae. Precipitation and temperature data are both first assessed
    by their Kendall and Spearman correlations. Next, they are passed through an AR(n) 
    filter -- default AR(1) -- to calculate residuals. These residuals can be visually 
    assessed for stationarity and their dependence structure (Kendall plots). 
    Pseudo-observations are created from the residuals, and then copulae of different 
    families are fit to the pseudo-observations (Independence, Frank, and Gaussian
    copula families are available in the current version).
    
    Parameters
    ----------
    data: pd.DataFrame
        Temporal reformat of ``raw_data`` where each year, month, day have their own column
        in the DataFrame
    resolution: str
        The temporal resolution of the input data. Can be 'monthly' or 'daily' 
    precip_fit_years: list[int]
        Years that precipitation was fit to
    ar_lag: int
        The time lag to consider in the AR fit step
    stationarity_groups: int
        The number of groups to consider in validating the stationarity of the residuals
    copula_families: list[str]
        The type of copula to consider when choosing a best-fitting family

    Returns
    -------
    pt_dict: dict
        Dictionary containing statistical information related to fitting of copulae/temp data
    """
     
    def investigate_autocorrelation(data_dict: dict, lag: int) -> dict:
        """
        Apply an autoregressive fit to the precipitation and temperature
        data to (1) find the fit, and; (2) calculate residuals

        Parameters
        ----------
        data_dict: dict
            Dictionary with the month-separated, spatially-averaged precipitation 
            and temperature data
        lag: int
            Lag to apply in the AR fit

        Returns
        -------
        data_dict: dict
            Same dictionary as above, with new keys for the fitted data and residuals
        """
        
        for month in data_dict.keys():
            p, t = data_dict[month]["PRECIP"], pt_dict[month]["TEMP"]
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=DeprecationWarning)
                warnings.simplefilter("ignore", category=FutureWarning)
                data_dict[month]["PRECIP ARFit"] = AutoReg(p, lags=[lag]).fit()
                data_dict[month]["TEMP ARFit"] = AutoReg(t, lags=[lag]).fit()

            # find the underlying univariate distribution of the residuals
            p_univariate, t_univariate = univariate.Univariate(), univariate.Univariate()
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                p_univariate.fit(np.array(data_dict[month]["PRECIP ARFit"].resid, dtype=float))
                t_univariate.fit(np.array(data_dict[month]["TEMP ARFit"].resid, dtype=float))
            data_dict[month]["PRECIP Resid Dist"] = p_univariate
            data_dict[month]["TEMP Resid Dist"] = t_univariate
        
        # validate ACF of the raw data and residual autoregression/autocorrelation
        if do_validation and ("copula" in validation_figures):
            validate_pt_acf(validation_dirpath, validation_extension, data_dict, lag)
        
        return data_dict

    
    def fit_copula_families(data_dict: dict, families: list[str]) -> dict:
        """
        Fit copulae to the residuals and rank the best fitting copula
        (per month) according to the families outlined in ``copula_families``

        Parameters
        ----------
        data_dict: dict
            Dictionary with the month-separated, spatially-averaged precipitation 
            and temperature data
        copula_families: list[str]
            Families to consider in the copula fitting process

        Returns
        -------
        data_dict: dict
            Same dictionary as above, with new keys for fitted copula
        """
        
        def CalculateEmpiricalCopulaCDFatPoint(pseudo_observations: np.array, point: float) -> np.array:
            n = pseudo_observations.shape[0]
            vec_Cn = np.full(shape=n, fill_value=0.)
            for k in range(n):
                U1, U2 = pseudo_observations[k, 0], pseudo_observations[k, 1]
                if U1 <= point[0] and U2 <= point[1]:
                    vec_Cn[k] = 1.
            return np.sum(vec_Cn) / n

        def BootstrapCopulaCVMandKS(pseudo_observations: np.array, theory_copula, from_df: bool = False) -> list[np.array, np.array]:
            n = pseudo_observations.shape[0]
            if from_df:
                column_names = theory_copula.to_dict()["columns"]

            # Cramér Von-Mises (S_n), Kolmogorov-Smirnov (T_n)
            Sn_elements = np.full(shape=n, fill_value=np.nan)
            Tn_elements = np.full(shape=n, fill_value=np.nan)
            for k in range(n):
                po = pseudo_observations[k, :]
                C_n = CalculateEmpiricalCopulaCDFatPoint(pseudo_observations, po)
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", category=DeprecationWarning)
                    Bstar_m = theory_copula.cdf(pd.DataFrame(data={column_names[0]: [po[0]], column_names[1]: [po[1]]})) if from_df else theory_copula.cdf(np.atleast_2d(po))
                Bstar_m = Bstar_m[0] if type(Bstar_m) in [list, np.ndarray] else Bstar_m
                Sn_elements[k] = (C_n - Bstar_m)**2.
                Tn_elements[k] = np.abs(C_n  - Bstar_m)
            S_n = np.sum(Sn_elements)
            T_n = np.max(Tn_elements)

            return S_n, T_n

        def FindBestCopula(copula_df: pd.DataFrame):
            aic_sorted, sn_sorted, tn_sorted = copula_df.copy().sort_values(by=["AIC"]), copula_df.copy().sort_values(by=["S_n"]), copula_df.copy().sort_values(by=["T_n"])

            # choosing the best copula
            winning_families = [aic_sorted.index.values[0], sn_sorted.index.values[0], tn_sorted.index.values[0]]
            n_best_families = len(set(winning_families))
            if n_best_families == 1 or n_best_families == 3:
                # -- if all three metrics are the lowest for a single family, use that one
                # -- if each family claims one lowest metric, just use AIC
                # ---- if AIC is at least 2 less each other family, that's a better model
                return [copula_df.at[aic_sorted.index.values[0], "Copula"], aic_sorted.index.values[0]]
            else:
                # pick the family where both S_n and T_n are winning if true, otherwise pick the winning AIC family
                if len({sn_sorted.index.values[0], tn_sorted.index.values[0]}) == 1:
                    return [copula_df.at[winning_families[1], "Copula"], winning_families[1]]
                else:
                    return [copula_df.at[winning_families[0], "Copula"], winning_families[0]]


        # include the pseudo-observations in the dictionary
        for month in data_dict.keys():
            nP, nT = len(data_dict[month]["PRECIP ARFit"].resid), len(data_dict[month]["TEMP ARFit"].resid)
            p_resids, t_resids = data_dict[month]["PRECIP ARFit"].resid, data_dict[month]["TEMP ARFit"].resid
            mask = ~(np.isnan(p_resids) | np.isnan(t_resids))
            data_dict[month]["PRECIP pObs"] = scipy.stats.rankdata(p_resids[mask], method="average") / (nP+1)
            data_dict[month]["TEMP pObs"] = scipy.stats.rankdata(t_resids[mask], method="average") / (nT+1)

        copula_fit_dict = {month: pd.DataFrame(data={"Copula": [None] * len(families),
                                                     "params": [np.nan] * len(families),
                                                     "AIC": [np.inf] * len(families),
                                                     "S_n": [[]] * len(families),
                                                     "T_n": [[]] * len(families)},
                                               index=families) for month in data_dict}
        for month in data_dict:
            pseudo_obs = np.array([data_dict[month]["PRECIP pObs"], data_dict[month]["TEMP pObs"]]).T

            if "Independence" in families:
                iCop = copulae.IndepCopula()
                copula_fit_dict[month].at["Independence", "Copula"] = iCop
                copula_fit_dict[month].at["Independence", "params"] = np.nan
                copula_fit_dict[month].at["Independence", "AIC"] = eval_measures.aic(llf=iCop.log_lik(pseudo_obs),
                                                                                     nobs=pseudo_obs.size, df_modelwc=np.array([]).size)
                iS_n, iT_n = BootstrapCopulaCVMandKS(pseudo_obs, iCop)
                copula_fit_dict[month].at["Independence", "S_n"] = iS_n
                copula_fit_dict[month].at["Independence", "T_n"] = iT_n

            if "Frank" in families:
                fCop = bivariate.Frank() 
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", category=DeprecationWarning)
                    fCop.fit(pseudo_obs)
                copula_fit_dict[month].at["Frank", "Copula"] = fCop
                copula_fit_dict[month].at["Frank", "params"] = fCop.theta
                copula_fit_dict[month].at["Frank", "AIC"] = eval_measures.aic(llf=np.sum(fCop.log_probability_density(pseudo_obs)),
                                                                              nobs=pseudo_obs.size, df_modelwc=np.array(fCop.theta).size)
                fS_n, fT_n = BootstrapCopulaCVMandKS(pseudo_obs, fCop)
                copula_fit_dict[month].at["Frank", "S_n"] = fS_n
                copula_fit_dict[month].at["Frank", "T_n"] = fT_n

            if "Gaussian" in families:
                pseudo_obs_df = pd.DataFrame(data=pseudo_obs, columns=["uP", "uT"])
                gCop = multivariate.GaussianMultivariate(distribution={"uP": univariate.UniformUnivariate(),
                                                                       "uT": univariate.UniformUnivariate()})
                gCop.fit(pseudo_obs_df)
                copula_fit_dict[month].at["Gaussian", "Copula"] = gCop
                copula_fit_dict[month].at["Gaussian", "params"] = gCop.correlation["uP"]["uT"]
                gCopulae = copulae.GaussianCopula()
                gCopulae.params = gCop.correlation["uP"]["uT"]
                copula_fit_dict[month].at["Gaussian", "AIC"] = eval_measures.aic(llf=gCopulae.log_lik(data=pseudo_obs_df.values, to_pobs=False),
                                                                                 nobs=pseudo_obs.size, df_modelwc=np.array(gCop.correlation["uP"]["uT"]).size)
                gS_n, gT_n = BootstrapCopulaCVMandKS(pseudo_obs, gCop, from_df=True)
                copula_fit_dict[month].at["Gaussian", "S_n"] = gS_n
                copula_fit_dict[month].at["Gaussian", "T_n"] = gT_n

            # add the copula families dictionary to the ptDict, for conciseness
            data_dict[month]["CopulaDF"] = copula_fit_dict[month]
            data_dict[month]["BestCopula"] = FindBestCopula(copula_fit_dict[month])

        return data_dict
    
    
    sites = sorted(set(data["SITE"].values))
    years = precip_fit_years
    full_years = [y for y in range(np.nanmin(years), np.nanmax(years)+1)]
    month_names = ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]

    # aggregate the data into monthly data (if not already)
    if "DAY" in data.columns:
        monthed_dict = {}
        for site in sites:
            site_idx = data["SITE"] == site
            for year in years:
                year_idx = data["YEAR"] == year
                for i in range(len(month_names)):
                    month_idx = data["MONTH"] == i+1
                    month_entry = data.loc[site_idx & year_idx & month_idx]
                    if month_entry.empty:
                        monthly_precip, monthly_temp = np.nan, np.nan
                    else:
                        monthly_precip = np.nansum(month_entry["PRECIP"].values)
                        monthly_temp = np.nanmean(month_entry["TEMP"].values)
                    monthed_dict[(site, year, month_names[i])] = [site, year, month_names[i], monthly_precip, monthly_temp]
        pt_df = pd.DataFrame().from_dict(monthed_dict, orient="index", columns=["SITE", "YEAR", "MONTH", "PRECIP", "TEMP"])
        pt_df.reset_index(drop=True, inplace=True)
    else:
        pt_df = data.copy()
        pt_df["MONTH"] = [month_names[i-1] for i in pt_df["MONTH"].values]

    # validate exploration of correlation between precipitation and temperature if prompted
    if do_validation and ("copula" in validation_figures):
        validate_explore_pt_dependence(validation_dirpath, validation_extension, pt_df, years)

    # establishing a dictionary, filling missing values from the group average
    pt_dict = {month: {"PRECIP": [], "TEMP": []} for month in month_names}
    for month in pt_dict:
        month_idx = pt_df["MONTH"] == month
        p, t = [], []
        for year in full_years:
            year_idx = pt_df["YEAR"] == year
            spatial_entry = pt_df.loc[month_idx & year_idx]
            p.append(np.nan if (spatial_entry.empty) or (np.all(np.isnan(spatial_entry["PRECIP"].values))) else np.nanmean(spatial_entry["PRECIP"].astype(float).values))
            t.append(np.nan if (spatial_entry.empty) or (np.all(np.isnan(spatial_entry["TEMP"].values))) else np.nanmean(spatial_entry["TEMP"].astype(float).values))
        p, t = np.array(p), np.array(t)
        p[np.isnan(p)], t[np.isnan(t)] = np.nanmean(p), np.nanmean(t)
        pt_dict[month]["PRECIP"], pt_dict[month]["TEMP"] = p, t
    
    # autocorrelation/autoregression of the precipitation and temperature data
    pt_dict = investigate_autocorrelation(pt_dict, ar_lag) 
   
    # stationarity of the precipitation and temperature residuals
    # -- note: Tootoonchi (2022) suggests that for exploratory applications
    # -- that stationarity is not necessarily something to correct for, and
    # -- can be considered a part of the conditional variability between the
    # -- two marginals. We take that approach below, so this function only
    # -- identifies/validates the existing stationarity of the residuals,
    # -- if there is any
    if do_validation and ("copula" in validation_figures):
        validate_pt_stationarity(validation_dirpath, validation_extension, pt_dict, stationarity_groups)

    # dependence structure of the residuals to help determining copula families
    if do_validation and ("copula" in validation_figures):
        validate_pt_dependence_structure(validation_dirpath, validation_extension, pt_dict)

    # fit the copula families
    copulaetemp_fit_dict = fit_copula_families(pt_dict, copula_families)

    return copulaetemp_fit_dict




import os
import numpy as np
import pandas as pd
import datetime as dt
import warnings

from .fit import fit_data
from .synthesize import synthesize_data


class SWXGModel:
    def __init__(self, raw_data: pd.DataFrame) -> None:
        """
        The base class to create, debias, fit, synthesize, and validate the stochastic 
        weather generation model.

        Parameters
        ----------
        raw_data: pd.DataFrame
            Input dataframe to use as for stochastic weather generation

        Attributes
        ----------
        raw_data: pd.DataFrame
            Mirrors in the input ``raw_data`` parameter
        data: pd.DataFrame
            Temporal reformatting of ``raw_data`` so that each year, month, day have their own column
            in the DataFrame
        resolution: str
            Determined resolution of the observed data. Can be "monthly" or "daily"
        precip_fit_dict: dict
            Dictionary containing statistical information related to fitting of precipitation data
        copulaetemp_fit_dict: dict
            Dictionary containing statistical information related to fitting of copulae and temperature data
        is_fit: bool:
            Flag to confirm that this instance of the generator has been fit
        """
        
        self.raw_data = raw_data
        self.resolution = ""
        self.data = pd.DataFrame()
        self.precip_fit_dict = {}
        self.copulaetemp_fit_dict = {}
        self.is_fit = False

        assert len(self.raw_data.columns) >= 4, "Input dataframe must have at least 4 columns!"
        assert ("SITE" in self.raw_data.columns) and (self.raw_data["SITE"].dtype == object), "Location ID column must be labeled 'SITE' with type 'str'!" 
        assert ("DATETIME" in self.raw_data.columns) and (self.raw_data["DATETIME"].dtype in [np.dtype('<M8[ns]'), np.dtype('>M8[ns]')]), "Date/Time column must be labeled 'DATETIME' with type datetime64[ns]!"
        assert ("PRECIP" in self.raw_data.columns) and (self.raw_data["PRECIP"].dtype == np.dtype('float64')), "Precipitation (with units [m]) column must be labeled 'PRECIP' with type 'float'!"
        assert ("TEMP" in self.raw_data.columns) and (self.raw_data["TEMP"].dtype == np.dtype('float64')), "Temperature (with units [degC]) column must be labeled 'TEMP' with type 'float'!"
        self.format_time_resolution(self.raw_data)


    def format_time_resolution(self, data: pd.DataFrame) -> None:
        """
        Function that separates the raw data's datetime stamps to individual dataframe 
        columns based on the input resolution
        
        Parameters
        ----------
        data: pd.DataFrame
            Input raw data to be used for the fitting
        """
        
        # determine the resolution
        days = [int(data.iloc[i]["DATETIME"].day) for i in range(data.shape[0])]
        resolution = "monthly" if len(set(days)) == 1 else "daily"
        subdaily_flag = False

        # define dataframe columns, datatypes
        if resolution == "monthly":
            stamp_cols = ["SITE", "YEAR", "MONTH", "PRECIP", *data.columns[3:]]
        else:
            stamp_cols = ["SITE", "YEAR", "MONTH", "DAY", "PRECIP", *data.columns[3:]]
        stamp_dtypes = {col: float for col in stamp_cols}
        stamp_dtypes["SITE"] = str
        stamp_dtypes["YEAR"], stamp_dtypes["MONTH"] = int, int
        if resolution == "daily":
            stamp_dtypes["DAY"] = int

        # separate dt.datetime column into years, months, (days)
        wvar_columns = ["PRECIP", *data.columns[3:]]
        dt_stamp_dict, i, subdaily_dict = {}, 0, {wvar: {} for wvar in wvar_columns}
        for i in range(data.shape[0]):
            df_row = data.iloc[i]
            site, year, month = df_row["SITE"], df_row["DATETIME"].year, df_row["DATETIME"].month
            precip = df_row["PRECIP"]
            temp_plus = [df_row[col] for col in data.columns[3:]]
            if resolution == "monthly":
                dt_stamp_dict[i] = [site, year, month, precip, *temp_plus]
            else:
                day = df_row["DATETIME"].day
                wvars = [precip, *temp_plus]
                for wv, wvar_col in enumerate(wvar_columns):
                    if (site, year, month, day) not in subdaily_dict[wvar_col]:
                        subdaily_dict[wvar_col][(site, year, month, day)] = [wvars[wv]] 
                    else:
                        subdaily_dict[wvar_col][(site, year, month, day)].append(wvars[wv])
                        subdaily_flag = True
                if not subdaily_flag:
                    dt_stamp_dict[i] = [site, year, month, day, precip, *temp_plus] 
        # if subdaily, aggregate/average to daily
        if subdaily_flag:
            for i, k in enumerate(subdaily_dict["PRECIP"].keys()):
                wvars = []
                for wvar_col in wvar_columns:
                    if wvar_col == "PRECIP": 
                        wvars.append(np.nansum(subdaily_dict[wvar_col][k]))
                    else:
                        wvars.append(np.nanmean(subdaily_dict[wvar_col][k]))
                dt_stamp_dict[i] = [*k, *wvars] 
        dt_stamp_df = pd.DataFrame().from_dict(dt_stamp_dict, orient="index", columns=stamp_cols)
        dt_stamp_df.reset_index(drop=True, inplace=True)
        dt_stamp_df.astype(stamp_dtypes) 

        self.data, self.resolution = dt_stamp_df, resolution
        if subdaily_flag:
            warnings.warn("Subdaily data resolution detected! Subdaily synthesizing not yet implemented, aggregated to daily...", UserWarning)
        if len(set(self.data["YEAR"].values)) < 20:
            warnings.warn("Fewer than 20 years of data detected! Fit and synthesizing is possible, but carefully review fit validation before synthesizing...", UserWarning)

    
    def fit(self, 
            verbose: bool = True,
            validate: bool = True,
            dirpath: str = "",
            kwargs: dict = {}) -> None:
        """
        Local method that calls the ``fit_data`` function, which aptly fits the input 
        weather data by returning extracted statistical parameters and reformatted data 

        Parameters
        ----------
        verbose: bool, optional
            Flag for displaying precipitation and temperature fit statistics. Default: True
        validate: bool, optional
            Flag for producing figures to validate each step of the generator. Default: True
        dirpath: str, optional
            Path for where to save the validation figures. Default: ""
        kwargs: dict, optional
            Keyword arguments related to the fit. Leaving this empty sets the keyword
            arguments to their default values. Keywords are:
            
             * ``gmmhmm_min_states``: int, default = 1
             * ``gmmhmm_max_states``: int, default = 4
             * ``gmmhmm_states``: int, default = 0
             * ``ar_lag``: int, default = 1
             * ``stationarity_groups``: int, default = 2
             * ``copula_families``: list[str], default = ["Independence", "Frank", "Gaussian"]
             * ``figure_extension``: str, default="svg"
             * ``validation_figures``: list[str], default = ["precip", "copula"]
        """
        
        kwargs["fit_verbose"] = verbose
        self.precip_fit_dict, self.copulaetemp_fit_dict = fit_data(self.data, self.resolution, validate, dirpath, kwargs)
        self.is_fit = True
        
        if verbose:
            print("--------------- Precipitation Fit ---------------")
            verbose_precip_sites, verbose_precip_cols = list(self.precip_fit_dict["log10_annual_precip"].columns), ["STATE", "SITE", "MEANS", "STDS"]
            verbose_gmmhmm_df = pd.DataFrame(columns=verbose_precip_cols)
            states, sites, means, stds = [], [], [], []
            for s in range(self.precip_fit_dict["num_gmmhmm_states"]):
                states.extend([s] * len(verbose_precip_sites))
                for i, site in enumerate(verbose_precip_sites):
                    sites.append(site)
                    means.append(self.precip_fit_dict["means"][s][i])
                    stds.append(self.precip_fit_dict["stds"][s][i])
            verbose_gmmhmm_df["STATE"], verbose_gmmhmm_df["SITE"], verbose_gmmhmm_df["MEANS"], verbose_gmmhmm_df["STDS"] = states, sites, means, stds
            verbose_transprob_idxs = ["FROM STATE {}".format(t) for t in range(self.precip_fit_dict["num_gmmhmm_states"])]
            verbose_transprob_cols = ["TO STATE {}".format(t) for t in range(self.precip_fit_dict["num_gmmhmm_states"])]
            verbose_transprob_df = pd.DataFrame(index=verbose_transprob_idxs, columns=verbose_transprob_cols) 
            for j in range(self.precip_fit_dict["num_gmmhmm_states"]):
                for k in range(self.precip_fit_dict["num_gmmhmm_states"]):
                    verbose_transprob_df.at["FROM STATE {}".format(j), "TO STATE {}".format(k)] = self.precip_fit_dict["t_probs"][j][k]
            print("* Number of GMMHMM States: {}".format(self.precip_fit_dict["num_gmmhmm_states"])) 
            print(" ")
            print("* GMMHMM Means/Stds per Site and State")
            print(verbose_gmmhmm_df.to_string(index=False))
            print(" ")
            print("* Transition Probability Matrix")
            print(verbose_transprob_df.to_string())
            print("-------------------------------------------------") 
            print(" ")
            print("------------------ Copulas Fit ------------------")
            months = self.copulaetemp_fit_dict.keys()
            for month in months:
                month_df = self.copulaetemp_fit_dict[month]
                print("Copula Statistics for: {}".format(month.upper()))
                print("* Best-Fitting Copula Family: {}".format(month_df["BestCopula"][1]))
                print("* All Family Parameters and Fit Comparison")
                families_df = month_df["CopulaDF"]
                families_df.drop("Copula", axis=1, inplace=True)
                families_df.rename(columns={"params": "Hyperparameter", "S_n": "Cram\u00e9r von Mises", "T_n": "Kolmogorov-Smirnov"}, inplace=True)
                print(families_df.to_string())
                print(" ")
            print("-------------------------------------------------") 

    
    def synthesize(self,
                   n: int = 0,
                   resolution: str = "",
                   validate: bool = True,
                   dirpath: str = "",
                   kwargs: dict = {}) -> pd.DataFrame:
        """
        Local method that calls the ``synthesize_data`` function, which aptly 
        synthesizes weather through either observed fit or given statistical parameters 

        Parameters
        ----------
        n: int, optional
            Number of years to synthesize. Default takes the same size as the number of
            years in the observed dataset
        resolution: str, optional
            The resolution to synthesize the data at. Leaving this empty sets the same
            resolution as the raw data
        validate: bool, optional
            Flag for producing figures to validate each step of the generator. Default: True
        dirpath: str, optional
            Path for where to save the validation figures. Default: ""
        kwargs: dict, optional
            Keyword arguments related to the fit. Leaving this empty sets the keyword
            arguments to their default values. Keywords are:
            
             * ``validation_samplesize_mult``: int, default = 10
             * ``figure_extension``: str, default="svg"

        Returns
        -------
        synthesized_data: pd.DataFrame
            The synthesized weather data at the desired time resolution
        """

        assert not self.data.empty, "Must include a dataframe of weather observations but none found!"
        assert self.is_fit, "Data has not been fit and therefore cannot synthesize!" 
        resolution = self.resolution if resolution == "" else resolution

        n = len(set(self.data["YEAR"].values)) if n <= 0 else n
        if ("DAY" in self.data) and (resolution == "monthly"):
            obs_dict = {}
            for site in sorted(set(self.data["SITE"].values)):
                site_idx = self.data["SITE"] == site
                site_entry = self.data.loc[site_idx]
                for year in sorted(set(site_entry["YEAR"].values)):
                    year_idx = site_entry["YEAR"] == year
                    year_entry = site_entry.loc[year_idx]
                    for month in sorted(set(year_entry["MONTH"].values)):
                        month_idx = year_entry["MONTH"] == month
                        month_entry = year_entry.loc[month_idx]
                        prcps, temps = month_entry["PRECIP"].values, month_entry["TEMP"].values
                        obs_dict[(site, year, month)] = [site, year, month, np.nansum(prcps), np.nanmean(temps)]
            obs_data = pd.DataFrame().from_dict(obs_dict, orient="index", columns=["SITE", "YEAR", "MONTH", "PRECIP", "TEMP"])
            obs_data.reset_index(drop=True, inplace=True)
            obs_data.astype({"SITE": str, "YEAR": int, "MONTH": int, "PRECIP": float, "TEMP": float})
        else:
            obs_data = self.data

        synthesized_data = synthesize_data(n, obs_data, self.precip_fit_dict, self.copulaetemp_fit_dict, 
                                           resolution, validate, dirpath, kwargs) 
        return synthesized_data



import os
import pandas as pd


monthly = pd.read_pickle(os.path.dirname(__file__) + "/example_monthly.pkl")
daily = pd.read_pickle(os.path.dirname(__file__) + "/example_daily.pkl")

__all__ = ["monthly", "daily"]




import os
from .model import SWXGModel
#from . import test_wx

#__all__ = ["SWXGModel", "test_wx"]
__all__ = ["SWXGModel"]
if not os.getenv("READTHEDOCS"):
    from . import test_wx
    __all__.append("test_wx")
